"""
A (mu+lambda) Memetic Algorithm using Randomized Local Search (RLS).

A memetic algorithm (:class:`~moptipy.algorithms.so.ma.MA`) works similar to a
(mu+lambda) :class:`~moptipy.algorithms.so.ea.EA`, but it refines all results
of the search operators with a local search, in this case,
:class:`~moptipy.algorithms.so.rls.RLS`, which is executed for
:attr:`~moptipy.algorithms.so.ma.MA.ls_fes` objective function evaluations.
It also only employs the nullary search operator (to create the initial random
solutions) and the binary search operator (to combine two selected parental
solutions), leaving the application of the unary search operator to the RLS.

A general implementation of a Memetic Algorithm into which arbitrary
algorithms can be plugged is given in :mod:`~moptipy.algorithms.so.ma`.
Here, the RLS part and the EA part of the MA are directly merged in a
hard-coded fashion. If the general :class:`~moptipy.algorithms.so.ma.MA` is
configured equivalently to the :class:`~moptipy.algorithms.so.marls.MARLS`
here, i.e., uses the same search operators, same `mu`, `lambda_`, and
`ls_fes`, then both algorithms will do exactly the same search steps.

1. Pablo Moscato. *On Evolution, Search, Optimization, Genetic Algorithms and
   Martial Arts: Towards Memetic Algorithms.* Caltech Concurrent Computation
   Program Report C3P 826. 1989. Pasadena, CA, USA: California Institute of
   Technology (Caltech), Caltech Concurrent Computation Program (C3P).
   https://www.researchgate.net/publication/2354457
2. Carlos Cotta, Luke Mathieson, and Pablo Moscato. Memetic Algorithms. In
   Rafael Martí, Panos M. Pardalos, and Mauricio G. C. Resende, editors,
   *Handbook of Heuristics.* Part~III: Metaheuristics, pages 607-638. 2018.
   Cham, Switzerland: Springer. ISBN: 978-3-319-07123-7.
   doi: https://doi.org/10.1007/978-3-319-07153-4_29-1
   https://www.researchgate.net/publication/315660932
3. William Eugene Hart, James E. Smith, and Natalio Krasnogor, editors.
   *Recent Advances in Memetic Algorithms.* Studies in Fuzziness and Soft
   Computing (STUDFUZZ), volume 166. 2005. Berlin/Heidelberg, Germany:
   Springer. ISBN: 978-3-540-22904-9.
   doi: https://doi.org/10.1007/3-540-32363-5
4. Ferrante Neri, Carlos Cotta, and Pablo Moscato. *Handbook of Memetic
   Algorithms.* Volume 379 of Studies in Computational Intelligence (SCI).
   2012. Berlin/Heidelberg, Germany: Springer. ISBN: 978-3-642-23246-6
   doi https://doi.org/10.1007/978-3-642-23247-3.
5. L. Darrell Whitley, V. Scott Gordon, and Keith E. Mathias. Lamarckian
   Evolution, The Baldwin Effect and Function Optimization. In Yuval Davidor,
   Hans-Paul Schwefel, and Reinhard Männer, editors, *Proceedings of the Third
   Conference on Parallel Problem Solving from Nature; International
   Conference on Evolutionary Computation (PPSN III),* October 9-14, 1994,
   Jerusalem, Israel, pages 5-15. Volume 866/1994 of Lecture Notes in Computer
   Science, Berlin, Germany: Springer-Verlag GmbH. ISBN 0387584846.
   doi: https://doi.org/10.1007/3-540-58484-6_245.
   https://www.researchgate.net/publication/2521727
"""
from typing import Any, Callable, Final, cast

from numpy.random import Generator

from moptipy.algorithms.so.record import Record
from moptipy.api.algorithm import Algorithm2
from moptipy.api.operators import Op0, Op1, Op2
from moptipy.api.process import Process
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.strings import PART_SEPARATOR
from moptipy.utils.types import check_int_range, type_error


# start book
class MARLS(Algorithm2):
    """
    A Memetic Algorithm as Evolutionary Algorithm + Randomized Local Search.

    This class implements a Memetic Algorithm (MA) basically as an
    Evolutionary Algorithm (EA), which only uses the nullary and binary search
    operators, and that refines each newly generated solution by applying a
    Randomized Local Search (RLS), which employs the unary search operator.
    """

    def solve(self, process: Process) -> None:
        """
        Apply the MA=EA+RLS to an optimization problem.

        :param process: the black-box process object
        """
        mu: Final[int] = self.mu  # mu: number of best solutions kept
        mu_plus_lambda: Final[int] = mu + self.lambda_  # size
        # initialization of some variables omitted in book for brevity
        # end book
        random: Final[Generator] = process.get_random()  # random gen
        create: Final[Callable] = process.create  # create x container
        evaluate: Final[Callable] = process.evaluate  # the objective
        op0: Final[Callable] = self.op0.op0  # the nullary operator
        op1: Final[Callable] = self.op1.op1  # the unary operator
        op2: Final[Callable] = self.op2.op2  # the binary operator
        ls_fes: Final[int] = self.ls_fes  # the rate at which to use op2
        should_terminate: Final[Callable] = process.should_terminate
        r0i: Final[Callable[[int], int]] = cast(  # random integers
            Callable[[int], int], random.integers)
        copy: Final[Callable[[Any, Any], None]] = process.copy

        # create list of mu random+ls records and lambda empty records
        # start book
        lst: Final[list] = [None] * mu_plus_lambda  # pre-allocate list
        f: int | float = 0  # variable to hold objective values  # -book
        tmp = create()  # the temporary record
        for i in range(mu_plus_lambda):  # fill list of size mu+lambda
            x = create()  # by creating point in search space
            if i < mu:  # only the first mu records are initialized by
                op0(random, x)  # applying nullary operator = randomize
                if should_terminate():  # should we quit?
                    return  # computational budget exhausted -> quit
                f = evaluate(x)  # continue? ok, evaluate new solution
                for _ in range(ls_fes):  # perform ls_fes of RLS
                    op1(random, tmp, x)  # unary operator
                    if should_terminate():  # should we quit?
                        return  # computational budget exhausted
                    ftmp: int | float = evaluate(tmp)  # evaluate new solution
                    if ftmp <= f:  # if it is better or equally good...
                        x, tmp = tmp, x  # accept it via swapping
                        f = ftmp  # and remember quality
            lst[i] = Record(x, f)  # create and store record

        it: int = 0  # set iteration counter=0 (but immediately increment)
        while True:  # lst: keep 0..mu-1, overwrite mu..mu+lambda-1
            it += 1  # step iteration counter
            for oi in range(mu, mu_plus_lambda):  # for all offspring
                if should_terminate():  # only continue if we still...
                    return  # have sufficient budget ... otherwise quit
                dest: Record = lst[oi]  # pick destination record
                x = dest.x  # the destination "x" value
                dest.it = it  # remember iteration of solution creation
                sx2 = sx = lst[r0i(mu)].x  # pick a random source "x"
                while sx2 is sx:  # until different from sx...
                    sx2 = lst[r0i(mu)].x  # ..get random second "x"
                op2(random, x, sx, sx2)  # apply binary operator
                f = evaluate(x)  # evaluate new point
                for _ in range(ls_fes):  # perform ls_fes of RLS
                    op1(random, tmp, x)  # unary operator
                    if should_terminate():  # should we quit?
                        return  # computational budget exhausted
                    ftmp = evaluate(tmp)  # evaluate new solution
                    if ftmp <= f:  # if it is better or equally good...
                        x, tmp = tmp, x  # accept it via swapping {HHH}
                        f = ftmp  # and remember quality
                if x is not dest.x:  # if we had swapped x away at {HHH}
                    copy(dest.x, x)  # store back solution
                    tmp = x  # and put x back into tmp variable
                dest.f = f  # store objective value of refined solution

            lst.sort()  # best records come first, ties broken by age
    # end book

    def __init__(self, op0: Op0,
                 op1: Op1,
                 op2: Op2,
                 mu: int = 1, lambda_: int = 1,
                 ls_fes: int = 1000,
                 name: str = "marls") -> None:
        """
        Create the Memetic Algorithm using hard-coded RLS as local search.

        :param op0: the nullary search operator
        :param op1: the unary search operator
        :param op2: the binary search operator
        :param mu: the number of best solutions to survive in each generation
        :param lambda_: the number of offspring in each generation
        :param ls_fes: the number of FEs (steps) per local search run
        :param name: the base name of the algorithm
        """
        if not isinstance(op0, Op0):
            raise type_error(op0, "op0", Op0)
        if not isinstance(op1, Op1):
            raise type_error(op1, "op1", Op1)
        if not isinstance(op2, Op2):
            raise type_error(op2, "op2", Op2)

        super().__init__(f"{name}{PART_SEPARATOR}{mu}{PART_SEPARATOR}"
                         f"{lambda_}{PART_SEPARATOR}{ls_fes}", op0, op1, op2)
        #: the number of records to survive in each generation
        self.mu: Final[int] = check_int_range(mu, "mu", 1, 1_000_000)
        #: the number of offsprings per generation
        self.lambda_: Final[int] = check_int_range(
            lambda_, "lambda", 1, 1_000_000)
        #: the number of FEs per local search run
        self.ls_fes: Final[int] = check_int_range(
            ls_fes, "ls_fes", 1, 100_000_000)

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of the algorithm to a logger.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value("mu", self.mu)
        logger.key_value("lambda", self.lambda_)
        logger.key_value("lsFEs", self.ls_fes)



"""Record for EndResult as well as parsing, serialization, and parsing."""
import argparse
import os.path
from dataclasses import dataclass
from math import inf, isfinite
from typing import Any, Callable, Final, Iterable

from moptipy.api.logging import (
    FILE_SUFFIX,
    KEY_ALGORITHM,
    KEY_BEST_F,
    KEY_GOAL_F,
    KEY_INSTANCE,
    KEY_LAST_IMPROVEMENT_FE,
    KEY_LAST_IMPROVEMENT_TIME_MILLIS,
    KEY_MAX_FES,
    KEY_MAX_TIME_MILLIS,
    KEY_RAND_SEED,
    KEY_TOTAL_FES,
    KEY_TOTAL_TIME_MILLIS,
    SECTION_FINAL_STATE,
    SECTION_SETUP,
)
from moptipy.evaluation._utils import (
    _FULL_KEY_GOAL_F,
    _FULL_KEY_MAX_FES,
    _FULL_KEY_MAX_TIME_MILLIS,
    _FULL_KEY_RAND_SEED,
    _check_max_time_millis,
)
from moptipy.evaluation.base import (
    F_NAME_NORMALIZED,
    F_NAME_RAW,
    F_NAME_SCALED,
    PerRunData,
)
from moptipy.evaluation.log_parser import ExperimentParser
from moptipy.utils.console import logger
from moptipy.utils.help import argparser
from moptipy.utils.logger import CSV_SEPARATOR, parse_key_values
from moptipy.utils.math import try_float_div, try_int
from moptipy.utils.path import Path
from moptipy.utils.strings import (
    intfloatnone_to_str,
    intnone_to_str,
    num_to_str,
    sanitize_names,
    str_to_intfloat,
    str_to_intfloatnone,
    str_to_intnone,
)
from moptipy.utils.types import check_int_range, type_error

#: The internal CSV header
_HEADER: Final[str] = f"{KEY_ALGORITHM}{CSV_SEPARATOR}" \
                      f"{KEY_INSTANCE}{CSV_SEPARATOR}" \
                      f"{KEY_RAND_SEED}{CSV_SEPARATOR}" \
                      f"{KEY_BEST_F}{CSV_SEPARATOR}" \
                      f"{KEY_LAST_IMPROVEMENT_FE}{CSV_SEPARATOR}" \
                      f"{KEY_LAST_IMPROVEMENT_TIME_MILLIS}" \
                      f"{CSV_SEPARATOR}" \
                      f"{KEY_TOTAL_FES}{CSV_SEPARATOR}" \
                      f"{KEY_TOTAL_TIME_MILLIS}{CSV_SEPARATOR}" \
                      f"{KEY_GOAL_F}{CSV_SEPARATOR}" \
                      f"{KEY_MAX_FES}{CSV_SEPARATOR}" \
                      f"{KEY_MAX_TIME_MILLIS}\n"


def __get_goal_f(e: "EndResult") -> int | float:
    """
    Get the goal_f.

    :param e: the end result
    :returns: the goal objective value
    """
    g = e.goal_f
    if g is None:
        raise ValueError(f"goal_f of {e} is None!")
    if not isfinite(g):
        raise ValueError(f"goal_f {g} of {e} is not finite!")
    return g


def __get_max_fes(e: "EndResult") -> int | float:
    """
    Get the max FEs.

    :param e: the end result
    :returns: the max fes
    """
    g = e.max_fes
    if g is None:
        raise ValueError(f"max_fes of {e} is None!")
    return g


def __get_max_time_millis(e: "EndResult") -> int | float:
    """
    Get the maximum time in milliseconds.

    :param e: the end result
    :returns: the maximum time in milliseconds
    """
    g = e.max_time_millis
    if g is None:
        raise ValueError(f"max_time_millis of {e} is None!")
    return g


def __get_goal_f_for_div(e: "EndResult") -> int | float:
    """
    Get the goal_f.

    :param e: the end result
    :returns: the goal objective value
    """
    g = __get_goal_f(e)
    if g <= 0:
        raise ValueError(f"goal_f {g} of {e}is not positive!")
    return g


def __get_f_norm(e: "EndResult") -> int | float:
    """
    Get the normalized f.

    :param e: the end result
    :returns: the normalized f
    """
    g = __get_goal_f_for_div(e)
    return try_float_div(e.best_f - g, g)


#: A set of getters for accessing variables of the end result
_GETTERS: Final[dict[str, Callable[["EndResult"], int | float]]] = {
    KEY_LAST_IMPROVEMENT_FE: lambda e: e.last_improvement_fe,
    KEY_LAST_IMPROVEMENT_TIME_MILLIS:
        lambda e: e.last_improvement_time_millis,
    KEY_TOTAL_FES: lambda e: e.total_fes,
    KEY_TOTAL_TIME_MILLIS: lambda e: e.total_time_millis,
    KEY_GOAL_F: __get_goal_f,
    F_NAME_RAW: lambda e: e.best_f,
    F_NAME_SCALED: lambda e: try_float_div(e.best_f, __get_goal_f_for_div(e)),
    F_NAME_NORMALIZED: __get_f_norm,
    KEY_MAX_FES: __get_max_fes,
    KEY_MAX_TIME_MILLIS: __get_max_time_millis,
}
_GETTERS[KEY_BEST_F] = _GETTERS[F_NAME_RAW]


@dataclass(frozen=True, init=False, order=True)
class EndResult(PerRunData):
    """
    An immutable end result record of one run of one algorithm on one problem.

    This record provides the information of the outcome of one application of
    one algorithm to one problem instance in an immutable way.
    """

    #: The best objective value encountered.
    best_f: int | float

    #: The index of the function evaluation when best_f was reached.
    last_improvement_fe: int

    #: The time when best_f was reached.
    last_improvement_time_millis: int

    #: The total number of performed FEs.
    total_fes: int

    #: The total time consumed by the run.
    total_time_millis: int

    #: The goal objective value if provided
    goal_f: int | float | None

    #: The (optional) maximum permitted FEs.
    max_fes: int | None

    #: The (optional) maximum runtime.
    max_time_millis: int | None

    def __init__(self,
                 algorithm: str,
                 instance: str,
                 rand_seed: int,
                 best_f: int | float,
                 last_improvement_fe: int,
                 last_improvement_time_millis: int,
                 total_fes: int,
                 total_time_millis: int,
                 goal_f: int | float | None,
                 max_fes: int | None,
                 max_time_millis: int | None):
        """
        Create a consistent instance of :class:`EndResult`.

        :param algorithm: the algorithm name
        :param instance: the instance name
        :param rand_seed: the random seed
        :param best_f: the best reached objective value
        :param last_improvement_fe: the FE when best_f was reached
        :param last_improvement_time_millis: the time when best_f was reached
        :param total_fes: the total FEs
        :param total_time_millis: the total runtime
        :param goal_f: the goal objective value, if provide
        :param max_fes: the optional maximum FEs
        :param max_time_millis: the optional maximum runtime

        :raises TypeError: if any parameter has a wrong type
        :raises ValueError: if the parameter values are inconsistent
        """
        super().__init__(algorithm, instance, rand_seed)
        object.__setattr__(self, "best_f", try_int(best_f))
        object.__setattr__(
            self, "last_improvement_fe", check_int_range(
                last_improvement_fe, "last_improvement_fe",
                1, 1_000_000_000_000_000))
        object.__setattr__(
            self, "last_improvement_time_millis", check_int_range(
                last_improvement_time_millis, "last_improvement_time_millis",
                0, 100_000_000_000))
        object.__setattr__(
            self, "total_fes", check_int_range(
                total_fes, "total_fes", last_improvement_fe,
                1_000_000_000_000_000))
        object.__setattr__(
            self, "total_time_millis", check_int_range(
                total_time_millis, "total_time_millis",
                last_improvement_time_millis, 100_000_000_000))

        if goal_f is not None:
            goal_f = None if goal_f <= -inf else try_int(goal_f)
        object.__setattr__(self, "goal_f", goal_f)

        if max_fes is not None:
            check_int_range(max_fes, "max_fes", total_fes, 1_000_000_000_000)
        object.__setattr__(self, "max_fes", max_fes)

        if max_time_millis is not None:
            check_int_range(
                max_time_millis, "max_time_millis", 1, 100_000_000_000)
            _check_max_time_millis(max_time_millis,
                                   total_fes,
                                   total_time_millis)
        object.__setattr__(self, "max_time_millis", max_time_millis)

    def success(self) -> bool:
        """
        Check if a run is successful.

        This method returns `True` if and only if `goal_f` is defined and
        `best_f <= goal_f` (and `False` otherwise).

        :return: `True` if and only if `best_f<=goal_f`
        """
        return False if self.goal_f is None else self.best_f <= self.goal_f

    def path_to_file(self, base_dir: str) -> Path:
        """
        Get the path that would correspond to the log file of this end result.

        Obtain a path that would correspond to the log file of this end
        result, resolved from a base directory `base_dir`.

        :param base_dir: the base directory
        :returns: the path to a file corresponding to the end result record
        """
        return Path.path(base_dir).resolve_inside(
            self.algorithm).resolve_inside(self.instance).resolve_inside(
            sanitize_names([self.algorithm, self.instance,
                            hex(self.rand_seed)]) + FILE_SUFFIX)

    @staticmethod
    def getter(dimension: str) -> Callable[["EndResult"], int | float]:
        """
        Produce a function that obtains the given dimension from EndResults.

        The following dimensions are supported:

        1. `lastImprovementFE`: :attr:`~EndResult.last_improvement_fe`
        2. `lastImprovementTimeMillis`:
            :attr:`~EndResult.last_improvement_time_millis`
        3. `totalFEs`: :attr:`~EndResult.total_fes`
        4. `totalTimeMillis`: :attr:`~EndResult.total_time_millis`
        5. `goalF`: :attr:`~EndResult.goal_f`
        6.  `plainF`, `bestF`: :attr:`~EndResult.best_f`
        7. `scaledF`: :attr:`~EndResult.best_f`/:attr:`~EndResult.goal_f`
        8. `normalizedF`: (:attr:`~EndResult.best_f`-attr:`~EndResult.goal_f`)/
            :attr:`~EndResult.goal_f`
        9. `maxFEs`: :attr:`~EndResult.max_fes`
        10. `maxTimeMillis`: :attr:`~EndResult.max_time_millis`
        11. `fesPerTimeMilli`:  :attr:`~EndResult.total_fes`
            /:attr:`~EndResult.total_time_millis`

        :param dimension: the dimension
        :returns: a callable that returns the value corresponding to the
            dimension from its input value, which must be an :class:`EndResult`
        """
        if not isinstance(dimension, str):
            raise type_error(dimension, "dimension", str)
        if dimension in _GETTERS:
            return _GETTERS[dimension]
        raise ValueError(f"unknown dimension {dimension!r}, "
                         f"should be one of {sorted(_GETTERS.keys())}.")

    @staticmethod
    def from_logs(path: str, consumer: Callable[["EndResult"], Any]) -> None:
        """
        Parse a given path and pass all end results found to the consumer.

        If `path` identifies a file with suffix `.txt`, then this file is
        parsed. The appropriate :class:`EndResult` is created and appended to
        the `collector`. If `path` identifies a directory, then this directory
        is parsed recursively for each log file found, one record is passed to
        the `consumer`. As `consumer`, you could pass any `callable` that
        accepts instances of :class:`EndResult`, e.g., the `append` method of
        a :class:`list`.

        :param path: the path to parse
        :param consumer: the consumer
        """
        _InnerLogParser(consumer).parse(path)

    @staticmethod
    def to_csv(results: Iterable["EndResult"], file: str) -> Path:
        """
        Write a sequence of end results to a file in CSV format.

        :param results: the end results
        :param file: the path
        :return: the path of the file that was written
        """
        path: Final[Path] = Path.path(file)
        logger(f"Writing end results to CSV file {path!r}.")
        Path.path(os.path.dirname(path)).ensure_dir_exists()

        with path.open_for_write() as out:
            out.write(_HEADER)
            for e in results:
                out.write(
                    f"{e.algorithm}{CSV_SEPARATOR}"
                    f"{e.instance}{CSV_SEPARATOR}"
                    f"{hex(e.rand_seed)}{CSV_SEPARATOR}"
                    f"{num_to_str(e.best_f)}{CSV_SEPARATOR}"
                    f"{e.last_improvement_fe}{CSV_SEPARATOR}"
                    f"{e.last_improvement_time_millis}{CSV_SEPARATOR}"
                    f"{e.total_fes}{CSV_SEPARATOR}"
                    f"{e.total_time_millis}{CSV_SEPARATOR}"
                    f"{intfloatnone_to_str(e.goal_f)}{CSV_SEPARATOR}"
                    f"{intnone_to_str(e.max_fes)}{CSV_SEPARATOR}"
                    f"{intnone_to_str(e.max_time_millis)}\n")

        logger(f"Done writing end results to CSV file {path!r}.")
        return path

    @staticmethod
    def from_csv(file: str, consumer: Callable[["EndResult"], Any],
                 filterer: Callable[["EndResult"], bool]
                 = lambda x: True) -> None:
        """
        Parse a given CSV file to get :class:`EndResult` Records.

        :param file: the path to parse
        :param consumer: the collector, can be the `append` method of a
            :class:`list`
        :param filterer: an optional filter function
        """
        if not callable(consumer):
            raise type_error(consumer, "consumer", call=True)
        path: Final[Path] = Path.file(file)
        logger(f"Now reading CSV file {path!r}.")

        with path.open_for_read() as rd:
            header = rd.readlines(1)
            if (header is None) or (len(header) <= 0):
                raise ValueError(f"No line in file {file!r}.")
            if header[0] != _HEADER:
                raise ValueError(
                    f"Header {header[0]!r} in {path!r} should be {_HEADER!r}.")

            while True:
                lines = rd.readlines(100)
                if (lines is None) or (len(lines) <= 0):
                    break
                for line in lines:
                    splt = line.strip().split(CSV_SEPARATOR)
                    er = EndResult(
                        splt[0].strip(),  # algorithm
                        splt[1].strip(),  # instance
                        int((splt[2])[2:], 16),  # rand seed
                        str_to_intfloat(splt[3]),  # best_f
                        int(splt[4]),  # last_improvement_fe
                        int(splt[5]),  # last_improvement_time_millis
                        int(splt[6]),  # total_fes
                        int(splt[7]),  # total_time_millis
                        str_to_intfloatnone(splt[8]),  # goal_f
                        str_to_intnone(splt[9]),  # max_fes
                        str_to_intnone(splt[10]))  # max_time_millis
                    if filterer(er):
                        consumer(er)

        logger(f"Done reading CSV file {path!r}.")


class _InnerLogParser(ExperimentParser):
    """The internal log parser class."""

    def __init__(self, consumer: Callable[[EndResult], Any]):
        """
        Create the internal log parser.

        :param consumer: the consumer accepting the parsed data
        """
        super().__init__()
        if not callable(consumer):
            raise type_error(consumer, "consumer", call=True)
        self.__consumer: Final[Callable[["EndResult"], Any]] = consumer
        self.__total_fes: int | None = None
        self.__total_time_millis: int | None = None
        self.__best_f: int | float | None = None
        self.__last_improvement_fe: int | None = None
        self.__last_improvement_time_millis: int | None = None
        self.__goal_f: int | float | None = None
        self.__max_fes: int | None = None
        self.__max_time_millis: int | None = None
        self.__state: int = 0

    def start_file(self, path: Path) -> bool:
        if not super().start_file(path):
            return False

        if self.__state != 0:
            raise ValueError(f"Illegal state when trying to parse {path}.")

        return True

    def end_file(self) -> bool:
        if self.__state != 3:
            raise ValueError(
                "Illegal state, log file must have both a "
                f"{SECTION_FINAL_STATE} and a "
                f"{SECTION_SETUP} section.")

        if self.rand_seed is None:
            raise ValueError("rand_seed is missing.")
        if self.algorithm is None:
            raise ValueError("algorithm is missing.")
        if self.instance is None:
            raise ValueError("instance is missing.")
        if self.__total_fes is None:
            raise ValueError("total_fes is missing.")
        if self.__total_time_millis is None:
            raise ValueError("total_time_millis is missing.")
        if self.__best_f is None:
            raise ValueError("best_f is missing.")
        if self.__last_improvement_fe is None:
            raise ValueError("last_improvement_fe is missing.")
        if self.__last_improvement_time_millis is None:
            raise ValueError("last_improvement_time_millis is missing.")

        self.__consumer(EndResult(self.algorithm,
                                  self.instance,
                                  self.rand_seed,
                                  self.__best_f,
                                  self.__last_improvement_fe,
                                  self.__last_improvement_time_millis,
                                  self.__total_fes,
                                  self.__total_time_millis,
                                  self.__goal_f,
                                  self.__max_fes,
                                  self.__max_time_millis))

        self.__total_fes = None
        self.__total_time_millis = None
        self.__best_f = None
        self.__last_improvement_fe = None
        self.__last_improvement_time_millis = None
        self.__goal_f = None
        self.__max_fes = None
        self.__max_time_millis = None
        self.__state = 0
        return super().end_file()

    def start_section(self, title: str) -> bool:
        super().start_section(title)
        if title == SECTION_SETUP:
            if (self.__state & 1) != 0:
                raise ValueError(f"Already did section {title}.")
            self.__state |= 4
            return True
        if title == SECTION_FINAL_STATE:
            if (self.__state & 2) != 0:
                raise ValueError(f"Already did section {title}.")
            self.__state |= 8
            return True
        return False

    def lines(self, lines: list[str]) -> bool:
        data = parse_key_values(lines)
        if not isinstance(data, dict):
            raise type_error(data, "data", dict)

        if (self.__state & 4) != 0:
            if _FULL_KEY_GOAL_F in data:
                goal_f = data[_FULL_KEY_GOAL_F]
                if ("e" in goal_f) or ("E" in goal_f) or ("." in goal_f):
                    self.__goal_f = float(goal_f)
                elif goal_f == "-inf":
                    self.__goal_f = None
                else:
                    self.__goal_f = int(goal_f)
            else:
                self.__goal_f = None

            if _FULL_KEY_MAX_FES in data:
                self.__max_fes = int(data[_FULL_KEY_MAX_FES])
            if _FULL_KEY_MAX_TIME_MILLIS in data:
                self.__max_time_millis = \
                    int(data[_FULL_KEY_MAX_TIME_MILLIS])

            seed_check = int(data[_FULL_KEY_RAND_SEED])
            if seed_check != self.rand_seed:
                raise ValueError(
                    f"Found seed {seed_check} in log file, but file name "
                    f"indicates seed {self.rand_seed}.")

            self.__state = (self.__state | 1) & (~4)
            return self.__state != 3

        if (self.__state & 8) != 0:
            self.__total_fes = int(data[KEY_TOTAL_FES])
            self.__total_time_millis = \
                int(data[KEY_TOTAL_TIME_MILLIS])

            self.__best_f = str_to_intfloat(data[KEY_BEST_F])

            self.__last_improvement_fe = \
                int(data[KEY_LAST_IMPROVEMENT_FE])
            self.__last_improvement_time_millis = \
                int(data[KEY_LAST_IMPROVEMENT_TIME_MILLIS])

            self.__state = (self.__state | 2) & (~8)
            return self.__state != 3

        raise ValueError("Illegal state.")


# Run log files to end results if executed as script
if __name__ == "__main__":
    parser: Final[argparse.ArgumentParser] = argparser(
        __file__,
        "Convert log files obtained with moptipy to the end results CSV "
        "format that can be post-processed or exported to other tools.",
        "This program recursively parses a folder hierarchy created by"
        " the moptipy experiment execution facility. This folder "
        "structure follows the scheme of algorithm/instance/log_file "
        "and has one log file per run. As result of the parsing, one "
        "CSV file (where columns are separated by ';') is created with"
        " one row per log file. This row contains the end-of-run state"
        " loaded from the log file. Whereas the log files may store "
        "the complete progress of one run of one algorithm on one "
        "problem instance as well as the algorithm configuration "
        "parameters, instance features, system settings, and the final"
        " results, the end results CSV file will only represent the "
        "final result quality, when it was obtained, how long the runs"
        " took, etc. This information is much denser and smaller and "
        "suitable for importing into other tools such as Excel or for "
        "postprocessing.")
    parser.add_argument(
        "source", nargs="?", default="./results",
        help="the location of the experimental results, i.e., the root folder "
             "under which to search for log files", type=Path.path)
    parser.add_argument(
        "dest", help="the path to the end results CSV file to be created",
        type=Path.path, nargs="?", default="./evaluation/end_results.txt")
    args: Final[argparse.Namespace] = parser.parse_args()

    end_results: Final[list[EndResult]] = []
    EndResult.from_logs(args.source, end_results.append)
    EndResult.to_csv(end_results, args.dest)



"""Test all the example code in the project's README.md file."""
import os.path
from typing import Final

from moptipy.utils.console import logger
from moptipy.utils.path import Path
from moptipy.utils.temp import TempDir


def test_all_examples_from_readme_md() -> None:
    """Test all the example Python codes in the README.md file."""
    # First, we load the README.md file as a single string
    base_dir = Path.directory(os.path.join(os.path.dirname(__file__), "../"))
    readme = Path.file(base_dir.resolve_inside("README.md"))
    logger(f"executing all examples from README.md file {readme!r}.")
    text = readme.read_all_str()
    logger(f"got {len(text)} characters.")
    if len(text) <= 0:
        raise ValueError(f"README.md file at {readme!r} is empty?")
    del readme

    i2: int = -1
    # All examples start and end with ``` after a newline.
    mark1: Final[str] = "\n```"
    mark2: Final[str] = "python"  # python code starts with ```python

    wd: Final[str] = os.getcwd()  # get current working directory
    # We run all the example codes in a temporary directory.
    with TempDir.create() as td:  # create temporary working directory
        logger(f"using temp directory {td!r}.")
        os.chdir(td)  # set it as working directory
        while True:
            # First, find the starting mark.
            i2 += 1
            i1 = text.find(mark1, i2)
            if i1 <= i2:
                break  # no starting mark anymore: done
            i1 += len(mark1)
            i2 = text.find(mark1, i1)
            if i2 <= i1:
                raise ValueError("No end mark for start mark?")

            fragment = text[i1:i2].strip()  # get the fragment
            if len(fragment) <= 0:
                raise ValueError("Empty fragment?")
            i2 += len(mark1)
            if fragment.startswith(mark2):  # it is a python fragment
                i3 = fragment.find("\n")
                if i3 < len(mark2):
                    raise ValueError("Did not find newline?")
                fragment = fragment[i3 + 1:].strip()
                if len(fragment) <= 0:
                    raise ValueError("Empty python fragment?")
                # OK, now we only have code left.
                logt = fragment[0:min(100, len(fragment))]\
                    .replace("\n", "\\n")
                logger(f"now processing fragment {logt}...")
                del logt

                logger("now compiling fragment.")
                code = compile(  # noqa # nosec
                    fragment, f"README.md:{i1}:{i2}",  # noqa # nosec
                    mode="exec")  # noqa # nosec
                logger("now executing fragment.")
                exec(code, {})  # execute file    # noqa # nosec
                logger("successfully executed example fragment.")

    os.chdir(wd)  # go back to current original directory
    logger("finished executing all examples from README.md.")



"""
Provides the base class for all components of the moptipy API.

All elements of the `moptipy` API inherit from
:class:`~moptipy.api.component.Component`. If you implement a new such
component, you can test it using the pre-defined unit test routine
:func:`~moptipy.tests.component.validate_component`.
"""
from typing import Final

from moptipy.api import logging
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.types import type_name_of


class Component:
    """The base class for all components of the moptipy API."""

    def __repr__(self):
        """
        Get the string representation of this object.

        :return: the value returned by :meth:`__str__`
        """
        return str(self)

    def __str__(self):
        """
        Get the default to-string implementation returns the class name.

        :returns: the class name of this component

        >>> print(Component())
        Component
        """
        s: Final[str] = type_name_of(self)
        i: Final[int] = s.rfind(".")
        if i > 0:
            return s[i + 1:]
        return s

    def initialize(self) -> None:
        """
        Initialize this component before a new run.

        Before every run of the optimization algorithm, its `initialize`
        method is called. The algorithm in turn must call all the `initialize`
        methods of all of its components.
        """

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of this component as key-value pairs.

        :param logger: the logger for the parameters

        >>> from moptipy.utils.logger import InMemoryLogger
        >>> with InMemoryLogger() as l:
        ...     with l.key_values("C") as kv:
        ...         Component().log_parameters_to(kv)
        ...     text = l.get_log()
        >>> text[-2]
        'class: moptipy.api.component.Component'
        >>> text[-3]
        'name: Component'
        >>> len(text)
        4
        """
        logger.key_value(logging.KEY_NAME, str(self))
        logger.key_value(logging.KEY_CLASS, type_name_of(self))



"""Progress data over a run."""
from dataclasses import dataclass
from math import inf, isfinite
from typing import Any, Callable, Final

import numpy as np

from moptipy.api import logging
from moptipy.evaluation._utils import _FULL_KEY_GOAL_F, _FULL_KEY_RAND_SEED
from moptipy.evaluation.base import (
    F_NAME_RAW,
    F_NAME_SCALED,
    TIME_UNIT_FES,
    TIME_UNIT_MILLIS,
    PerRunData,
    check_f_name,
    check_time_unit,
)
from moptipy.evaluation.log_parser import ExperimentParser
from moptipy.utils.console import logger
from moptipy.utils.logger import (
    COMMENT_CHAR,
    CSV_SEPARATOR,
    KEY_VALUE_SEPARATOR,
    parse_key_values,
)
from moptipy.utils.nputils import is_all_finite, is_np_float, is_np_int
from moptipy.utils.path import Path
from moptipy.utils.strings import num_to_str, str_to_intfloat
from moptipy.utils.types import type_error


@dataclass(frozen=True, init=False, order=True)
class Progress(PerRunData):
    """An immutable record of progress information over a single run."""

    #: The time axis data.
    time: np.ndarray

    #: The unit of the time axis.
    time_unit: str

    #: The objective value data.
    f: np.ndarray

    #: the name of the objective value axis.
    f_name: str

    #: the standard value of the objective dimension.
    #: If :attr:`f_name` is `F_NAME_SCALED` or `F_NAME_NORMALIZED`.
    #: then this value has been used to normalize the data.
    f_standard: int | float | None

    def __init__(self,
                 algorithm: str,
                 instance: str,
                 rand_seed: int,
                 time: np.ndarray,
                 time_unit: str,
                 f: np.ndarray,
                 f_name: str,
                 f_standard: int | float | None = None,
                 only_improvements: bool = True):
        """
        Create a consistent instance of :class:`EndResult`.

        :param algorithm: the algorithm name
        :param instance: the instance name
        :param rand_seed: the random seed
        :param time: the time axis data
        :param time_unit: the unit of the time axis
        :param f: the objective value axis data
        :param f_name: the name of the objective value axis data
        :param f_standard: the value used to standardize of the objective
            value dimension
        :param only_improvements: enforce that f-values should be
            improving and time values increasing
        """
        super().__init__(algorithm, instance, rand_seed)

        if not isinstance(time, np.ndarray):
            raise type_error(time, "time data", np.ndarray)
        time.flags.writeable = False
        if len(time.shape) != 1:
            raise ValueError("time array must be one-dimensional, but "
                             f"has shape {time.shape}.")
        if not is_np_int(time.dtype):
            raise TypeError("time data must be integer-valued, "
                            f"but has type {time.dtype}.")
        tl = time.size
        if tl <= 0:
            raise ValueError("time data must not be empty.")
        if tl > 1:
            if only_improvements:
                if np.any(time[1:] <= time[:-1]):
                    raise ValueError("time data must be strictly increasing,"
                                     f"but encountered {time}.")
            elif np.any(time[1:] < time[:-1]):
                raise ValueError("time data must be monotonously"
                                 f"increasing, but encountered {time}.")

        object.__setattr__(self, "time", time)
        object.__setattr__(self, "time_unit", check_time_unit(time_unit))

        mintime = 1 if time_unit == TIME_UNIT_FES else 0
        if any(time < mintime):
            raise ValueError(f"No time value can be less than {mintime} if"
                             f" time unit is {time_unit}.")

        if not isinstance(f, np.ndarray):
            raise type_error(f, "f data", np.ndarray)
        f.flags.writeable = False
        if len(f.shape) != 1:
            raise ValueError(
                f"f array must be one-dimensional, but has shape {f.shape}.")
        if is_np_float(f.dtype):
            if not is_all_finite(f):
                raise ValueError("f must be all finite.")
        elif not is_np_int(f.dtype):
            raise TypeError("f data must be integer- or float valued, but"
                            f" encountered an {type(f)} of {f.dtype}.")
        fl = f.size
        if fl <= 0:
            raise ValueError("f data must not be empty.")
        if fl != tl:
            raise ValueError(f"Length {fl} of f data and length {tl} of "
                             "time data must be the same.")
        if not isinstance(only_improvements, bool):
            raise type_error(only_improvements, "only_improvements", bool)
        if only_improvements and (fl > 1):
            if np.any(f[1:-1] >= f[:-2]):
                raise ValueError(
                    "f data must be strictly decreasing, with "
                    "only the entry being permitted as exception.")
            if f[-1] > f[-2]:
                raise ValueError(f"last f-value ({f[-1]}) cannot be greater"
                                 f"than second-to-last ({f[-2]}).")
        object.__setattr__(self, "f", f)
        object.__setattr__(self, "f_name", check_f_name(f_name))

        if (f_name != F_NAME_RAW) and (f_standard is None):
            raise ValueError(f"If f_name is {F_NAME_RAW}, "
                             f"then f_standard cannot be {f_standard}.")
        if f_standard is not None:
            if isinstance(f_standard, float):
                if not isfinite(f_standard):
                    raise ValueError(f"f_standard cannot be {f_standard}.")
            elif not isinstance(f_standard, int):
                raise type_error(f_standard, "f_standard", (int, float))
        object.__setattr__(self, "f_standard", f_standard)

    @staticmethod
    def from_logs(path: str,
                  consumer: Callable[["Progress"], Any],
                  time_unit: str = TIME_UNIT_FES,
                  f_name: str = F_NAME_RAW,
                  f_standard: dict[str, int | float] | None = None,
                  only_improvements: bool = True) -> None:
        """
        Parse a given path and pass all progress data found to the consumer.

        If `path` identifies a file with suffix `.txt`, then this file is
        parsed. The appropriate :class:`Progress` is created and appended to
        the `collector`. If `path` identifies a directory, then this directory
        is parsed recursively for each log file found, one record is passed to
        the `consumer`. The `consumer` is simply a callable function. You could
        pass in the `append` method of a :class:`list`.

        :param path: the path to parse
        :param consumer: the consumer, can be the `append` method of a
            :class:`list`
        :param time_unit: the time unit
        :param f_name: the objective name
        :param f_standard: a dictionary mapping instances to standard values
        :param only_improvements: enforce that f-values should be improving and
            time values increasing
        """
        _InnerLogParser(time_unit, f_name, consumer, f_standard,
                        only_improvements).parse(path)

    def to_csv(self, file: str,
               put_header: bool = True) -> str:
        """
        Store a :class:`Progress` record in a CSV file.

        :param file: the file to generate
        :param put_header: should we put a header with meta-data?
        :return: the fully resolved file name
        """
        path: Final[Path] = Path.path(file)
        logger(f"Writing progress object to CSV file {path!r}.")

        with path.open_for_write() as out:
            sep: Final[str] = CSV_SEPARATOR
            if put_header:
                kv: Final[str] = KEY_VALUE_SEPARATOR
                cmt: Final[str] = COMMENT_CHAR
                out.write(
                    f"{cmt} {logging.KEY_ALGORITHM}{kv}{self.algorithm}\n")
                out.write(
                    f"{cmt} {logging.KEY_INSTANCE}{kv}{self.instance}\n")
                out.write(f"{cmt} {logging.KEY_RAND_SEED}{kv}"
                          f"{hex(self.rand_seed)}\n")
                if self.f_standard is not None:
                    out.write(
                        f"{cmt} {logging.KEY_GOAL_F}{kv}{self.f_standard}\n")
            out.write(f"{self.time_unit}{sep}{self.f_name}\n")
            for i, t in enumerate(self.time):
                out.write(f"{t}{sep}{num_to_str(self.f[i])}\n")

        logger(f"Done writing progress object to CSV file {path!r}.")

        path.enforce_file()
        return path


class _InnerLogParser(ExperimentParser):
    """The internal log parser class."""

    # States are OR combinations of the following:
    # --> 0 == initial and end state
    # --> 7 == all necessary data should have been collected
    #  8: SECTION_SETUP is in progress
    # 16: SECTION_FINAL_STATE is in progress
    # 32: SECTION_PROGRESS is in progress

    def __init__(self, time_unit: str, f_name: str,
                 consumer: Callable[[Progress], Any],
                 f_standard: dict[str, int | float] | None = None,
                 only_improvements: bool = True):
        """
        Create the internal log parser.

        :param time_unit: the time unit
        :param f_name: the objective name
        :param consumer: the consumer
        :param f_standard: a dictionary mapping instances to standard values
        :param only_improvements: enforce that f-values should be improving
            and time values increasing
        """
        super().__init__()
        if not callable(consumer):
            raise type_error(consumer, "consumer", call=True)
        self.__consumer: Final[Callable[[Progress], Any]] = consumer
        self.__time_unit = check_time_unit(time_unit)
        self.__f_name = check_f_name(f_name)
        self.__total_time: int | None = None
        self.__total_fes: int | None = None
        self.__last_fe: int | None = None
        self.__goal_f: int | float | None = None
        self.__t_collector: Final[list[int]] = []
        self.__f_collector: Final[list[int | float]] = []
        if not isinstance(only_improvements, bool):
            raise type_error(only_improvements, "only_improvements", bool)
        self.__only_improvements = only_improvements
        if (f_standard is not None) and (not isinstance(f_standard, dict)):
            raise type_error(f_standard, "f_standard", dict)
        self.__f_standard: Final[None | dict[str, int | float]] \
            = f_standard
        self.__state = 0

    def start_file(self, path: Path) -> bool:
        if not super().start_file(path):
            return False
        if self.__state != 0:
            raise ValueError(f"Illegal state when trying to parse {path}.")
        return True

    def end_file(self) -> bool:
        if self.__state != 7:
            raise ValueError(
                "Illegal state, log file must have a "
                f"{logging.SECTION_FINAL_STATE}, a "
                f"{logging.SECTION_SETUP}, and a "
                f"{logging.SECTION_PROGRESS} section.")

        if self.rand_seed is None:
            raise ValueError("rand_seed is missing.")
        if self.algorithm is None:
            raise ValueError("algorithm is missing.")
        if self.instance is None:
            raise ValueError("instance is missing.")
        if self.__total_time is None:
            raise ValueError("total time is missing.")
        if self.__total_fes is None:
            raise ValueError("total FEs are missing.")
        if not self.__f_collector:
            raise ValueError("f-collector cannot be empty.")
        if not self.__t_collector:
            raise ValueError("time-collector cannot be empty.")

        f_standard: int | float | None = None
        if (self.__f_standard is not None) and \
                (self.instance in self.__f_standard):
            f_standard = self.__f_standard[self.instance]
        if f_standard is None:
            f_standard = self.__goal_f
        if (self.__f_name != F_NAME_RAW) and (f_standard is None):
            raise ValueError(f"f_standard cannot be {f_standard} if f_name "
                             f"is {self.__f_name}.")

        tt = self.__total_time if (self.__time_unit == TIME_UNIT_MILLIS) \
            else self.__total_fes
        if tt < self.__t_collector[-1]:
            raise ValueError(
                f"Last time units {tt} inconsistent with last"
                f"recorded time unit {self.__t_collector[-1]}.")
        if self.__last_fe < self.__total_fes:
            if tt > self.__t_collector[-1]:
                self.__t_collector.append(tt)
                self.__f_collector.append(self.__f_collector[-1])
        elif self.__last_fe > self.__total_fes:
            raise ValueError(
                f"Last FE {self.__last_fe} inconsistent with total number"
                f"{self.__total_fes} of FEs.")

        ff: np.ndarray
        if self.__f_name == F_NAME_RAW:
            ff = np.array(self.__f_collector)
        elif self.__f_name == F_NAME_SCALED:
            ff = np.array([f / f_standard for f in self.__f_collector])
        else:
            ff = np.array([(f - f_standard) / f_standard
                           for f in self.__f_collector])
        self.__f_collector.clear()

        self.__consumer(Progress(self.algorithm,
                                 self.instance,
                                 self.rand_seed,
                                 np.array(self.__t_collector),
                                 self.__time_unit,
                                 ff,
                                 self.__f_name,
                                 f_standard,
                                 self.__only_improvements))

        self.__total_time = None
        self.__goal_f = None
        self.__t_collector.clear()
        self.__total_time = None
        self.__total_fes = None
        self.__last_fe = None
        self.__state = 0
        return super().end_file()

    def start_section(self, title: str) -> bool:
        super().start_section(title)
        if title == logging.SECTION_SETUP:
            if (self.__state & 1) != 0:
                raise ValueError(f"Already did section {title}.")
            self.__state |= 8
            return True
        if title == logging.SECTION_FINAL_STATE:
            if (self.__state & 2) != 0:
                raise ValueError(f"Already did section {title}.")
            self.__state |= 16
            return True
        if title == logging.SECTION_PROGRESS:
            if (self.__state & 4) != 0:
                raise ValueError(f"Already did section {title}.")
            self.__state |= 32
            return True
        return False

    def lines(self, lines: list[str]) -> bool:
        if not isinstance(lines, list):
            raise type_error(lines, "lines", list)

        if (self.__state & 24) != 0:  # final state or setup
            data = parse_key_values(lines)

            if not isinstance(data, dict):
                raise ValueError("Error when parsing data.")

            if (self.__state & 8) != 0:  # state
                if _FULL_KEY_GOAL_F in data:
                    goal_f = data[_FULL_KEY_GOAL_F]
                    if ("e" in goal_f) or ("E" in goal_f) or ("." in goal_f):
                        self.__goal_f = float(goal_f)
                    elif goal_f == "-inf":
                        self.__goal_f = None
                    else:
                        self.__goal_f = int(goal_f)
                else:
                    self.__goal_f = None

                seed_check = int(data[_FULL_KEY_RAND_SEED])
                if seed_check != self.rand_seed:
                    raise ValueError(
                        f"Found seed {seed_check} in log file, but file name "
                        f"indicates seed {self.rand_seed}.")

                self.__state = (self.__state | 1) & (~8)
                return self.__state != 7

            if (self.__state & 16) != 0:  # END_STATE
                self.__total_fes = int(data[logging.KEY_TOTAL_FES])
                self.__total_time = \
                    int(data[logging.KEY_TOTAL_TIME_MILLIS])
                self.__state = (self.__state | 2) & (~16)
                return self.__state != 7

        if (self.__state & 32) != 0:  # CSV data
            n_rows = len(lines)
            if n_rows < 2:
                raise ValueError("lines must contain at least two elements,"
                                 f"but contains {n_rows}.")

            columns = [c.strip() for c in
                       lines[0].split(CSV_SEPARATOR)]
            n_cols = len(columns)
            if n_cols < 3:
                raise ValueError("There must be at least three columns, "
                                 f"but found {n_cols} in {lines[0]!r}.")

            time_col_name: str = logging.PROGRESS_TIME_MILLIS if \
                self.__time_unit == TIME_UNIT_MILLIS else logging.PROGRESS_FES
            time_col_idx: int = -1
            f_col_idx: int = -1
            fe_col_idx: int = -1
            for idx, col in enumerate(columns):  # find the columns we
                if col == logging.PROGRESS_FES:
                    fe_col_idx = idx
                if col == time_col_name:
                    if time_col_idx >= 0:
                        raise ValueError(f"Time column {time_col_name} "
                                         "appears twice.")
                    time_col_idx = idx
                elif col == logging.PROGRESS_CURRENT_F:
                    if f_col_idx >= 0:
                        raise ValueError(
                            f"F column {logging.PROGRESS_CURRENT_F} "
                            "appears twice.")
                    f_col_idx = idx

            def aa(splt):  # noqa
                return splt[time_col_idx], splt[f_col_idx]

            time, f = zip(*[[c.strip()
                             for c in aa(line.split(CSV_SEPARATOR))]
                            for line in lines[1:]], strict=True)
            time = [int(t) for t in time]
            f = [str_to_intfloat(v) for v in f]
            if self.__only_improvements:
                biggest_t: int = -1
                best_f: int | float = inf
                for idx, t in enumerate(time):
                    v = f[idx]
                    if t > biggest_t:
                        if biggest_t >= 0:
                            self.__t_collector.append(biggest_t)
                            self.__f_collector.append(best_f)
                        best_f = v
                        biggest_t = t
                    elif v < best_f:
                        best_f = v
                if biggest_t >= 0:
                    self.__t_collector.append(biggest_t)
                    self.__f_collector.append(best_f)
            else:
                self.__t_collector.extend(time)
                self.__f_collector.extend(f)

            self.__last_fe = int((lines[-1].split(CSV_SEPARATOR))[fe_col_idx])
            if self.__last_fe <= 0:
                raise ValueError(f"Last FE cannot be {self.__last_fe}.")

            self.__state = (self.__state | 4) & (~32)
            return self.__state != 7

        raise ValueError("Illegal state.")



"""
A Python implementation of the W-Model benchmark problem.

This is a tunable benchmark problem for bit-string based search spaces.
It exhibits ruggedness, epistasis, deceptiveness, and (uniform) neutrality in
a tunable fashion.
In [1], a set of 19 diverse instances of the W-Model are selected for
algorithm benchmarking. These instances are provided via
:meth:`~moptipy.examples.bitstrings.w_model.WModel.default_instances`.

1. Thomas Weise, Yan Chen, Xinlu Li, and Zhize Wu. Selecting a diverse set of
   benchmark instances from a tunable model problem for black-box discrete
   optimization algorithms. *Applied Soft Computing Journal (ASOC)*, 92:106269,
   June 2020. doi: https://doi.org/10.1016/j.asoc.2020.106269
2. Thomas Weise and Zijun Wu. Difficult Features of Combinatorial Optimization
   Problems and the Tunable W-Model Benchmark Problem for Simulating them. In
   *Black Box Discrete Optimization Benchmarking (BB-DOB) Workshop* of
   *Companion Material Proceedings of the Genetic and Evolutionary Computation
   Conference (GECCO 2018),* July 15th-19th 2018, Kyoto, Japan,
   pages 1769-1776, ISBN 978-1-4503-5764-7. ACM.
   doi: https://doi.org/10.1145/3205651.3208240
3. Thomas Weise, Stefan Niemczyk, Hendrik Skubch, Roland Reichle, and
   Kurt Geihs. A Tunable Model for Multi-Objective, Epistatic, Rugged, and
   Neutral Fitness Landscapes. In Maarten Keijzer, Giuliano Antoniol,
   Clare Bates Congdon, Kalyanmoy Deb, Benjamin Doerr, Nikolaus Hansen,
   John H. Holmes, Gregory S. Hornby, Daniel Howard, James Kennedy,
   Sanjeev P. Kumar, Fernando G. Lobo, Julian Francis Miller, Jason H. Moore,
   Frank Neumann, Martin Pelikan, Jordan B. Pollack, Kumara Sastry,
   Kenneth Owen Stanley, Adrian Stoica, El-Ghazali, and Ingo Wegener, editors,
   *Proceedings of the 10th Annual Conference on Genetic and Evolutionary
   Computation (GECCO'08)*, pages 795-802, July 12-16, 2008, Atlanta, GA, USA.
   ISBN: 978-1-60558-130-9, New York, NY, USA: ACM Press.
   doi: https://doi.org/10.1145/1389095.1389252
4. Carola Doerr, Furong Ye, Naama Horesh, Hao Wang, Ofer M. Shir, Thomas Bäck.
   Benchmarking Discrete Optimization Heuristics with IOHprofiler. *Applied
   Soft Computing* 88:106027, March 2020,
   doi: https://doi.org/10.1016/j.asoc.2019.106027.
"""

from array import array
from math import sqrt
from typing import Callable, Final, Iterable, cast

import numba  # type: ignore
import numpy as np

from moptipy.examples.bitstrings.bitstring_problem import BitStringProblem
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.nputils import DEFAULT_BOOL
from moptipy.utils.strings import sanitize_name
from moptipy.utils.types import check_int_range, type_error


@numba.njit(nogil=True, cache=True, inline="always")
def w_model_f(x: np.ndarray) -> int:
    """
    Compute the basic hamming distance of the W-Model to the optimal string.

    The optimal string in the W-Model is `0101010101...`. Here we compute the
    objective value of a candidate solution, i.e., the Hamming distance to
    the string `010101010101...`. This is the basic problem objective
    function. It can be applied either directly, on top of transformations
    such as the neutrality- or epistasis mappings. Its result can be
    transformed by a ruggedness permutation to introduce ruggedness into the
    problem.

    :param x: the bit string to evaluate
    :return: the Hamming distance to the string of alternating zeros and ones
        and starting with `0`.

    >>> w_model_f(np.array([False]))
    0
    >>> w_model_f(np.array([True]))
    1
    >>> w_model_f(np.array([False, True]))
    0
    >>> w_model_f(np.array([False, False]))
    1
    >>> w_model_f(np.array([True, True]))
    1
    >>> w_model_f(np.array([True, False]))
    2
    >>> w_model_f(np.array([False, True, False]))
    0
    >>> w_model_f(np.array([True, False, True]))
    3
    >>> w_model_f(np.array([True, True, True]))
    2
    >>> w_model_f(np.array([True, True, False]))
    1
    >>> w_model_f(np.array([False, True, True]))
    1
    >>> w_model_f(np.array([False, True, False, True, False, True, False]))
    0
    >>> w_model_f(np.array([False, True, False, True, False, True, True]))
    1
    >>> w_model_f(np.array([True, False, True, False, True, False, True]))
    7
    """
    result = 0
    for i, xx in enumerate(x):
        if xx == ((i & 1) == 0):
            result = result + 1
    return result


@numba.njit(nogil=True, cache=True, inline="always")
def w_model_neutrality(x_in: np.ndarray, mu: int, x_out: np.ndarray) -> None:
    """
    Perform the neutrality transformation.

    The neutrality transformation is the first layer of the W-Model. It
    introduces (or better, removes during the mapping) uniform redundancy by
    basically reducing the size of a bit string by factor `mu`.

    Basically, the input array `x_in` is split in consecutive bit groups of
    length `mu`. Each such group is translated to one bit in the output string
    `x_out`. This bit will become `1` if the majority of the bits in the input
    group are also `1`. Otherwise it becomes `0`.

    Notice that `len(x_out) >= mu * length(x_in)` must hold.

    :param x_in: the input array
    :param mu: the number of bits to merge
    :param x_out: the output array

    >>> xin = np.array([0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0,
    ...                 0, 1, 1, 1, 0, 1, 0, 0, 0, 0], bool)
    >>> xout = np.empty(len(xin) // 2, bool)
    >>> w_model_neutrality(xin, 2, xout)
    >>> print("".join('1' if xx else '0' for xx in xout))
    1011001110
    >>> xin = np.array([0, 1, 0, 1, 0, 1, 1, 0, 0, 0, 0,
    ...                 0, 1, 1, 1, 0, 1, 0, 0, 0, 0], bool)
    >>> w_model_neutrality(xin, 2, xout)
    >>> print("".join('1' if xx else '0' for xx in xout))
    1111001110
    """
    threshold_for_1: Final[int] = (mu >> 1) + (mu & 1)
    flush: int = mu
    i_in: int = 0
    i_out: int = 0
    ones: int = 0
    length_in: Final[int] = len(x_in)
    length_out: Final[int] = len(x_out)
    while (i_in < length_in) and (i_out < length_out):
        if x_in[i_in]:
            ones = ones + 1
        i_in = i_in + 1
        if i_in >= flush:
            flush = flush + mu
            x_out[i_out] = (ones >= threshold_for_1)
            i_out = i_out + 1
            ones = 0


@numba.njit(nogil=True, cache=True, inline="always")
def __w_model_epistasis(x_in: np.ndarray, start: int,
                        nu: int, x_out: np.ndarray) -> None:
    """
    Perform the transformation of an epistasis chunk.

    :param x_in: the input string
    :param start: the start index
    :param nu: the epistasis parameter
    :param x_out: the output string

    >>> xout = np.empty(4, bool)
    >>> __w_model_epistasis(np.array([0, 0, 0, 0], bool), 0, 4, xout)
    >>> print("".join('1' if xx else '0' for xx in xout))
    0000
    >>> __w_model_epistasis(np.array([0, 0, 0, 1], bool), 0, 4, xout)
    >>> print("".join('1' if xx else '0' for xx in xout))
    1101
    >>> __w_model_epistasis(np.array([0, 0, 1, 0], bool), 0, 4, xout)
    >>> print("".join('1' if xx else '0' for xx in xout))
    1011
    >>> __w_model_epistasis(np.array([0, 1, 0, 0], bool), 0, 4, xout)
    >>> print("".join('1' if xx else '0' for xx in xout))
    0111
    >>> __w_model_epistasis(np.array([1, 0, 0, 0], bool), 0, 4, xout)
    >>> print("".join('1' if xx else '0' for xx in xout))
    1111
    >>> __w_model_epistasis(np.array([1, 1, 1, 1], bool), 0, 4, xout)
    >>> print("".join('1' if xx else '0' for xx in xout))
    1110
    >>> __w_model_epistasis(np.array([0, 1, 1, 1], bool), 0, 4, xout)
    >>> print("".join('1' if xx else '0' for xx in xout))
    0001
    >>> __w_model_epistasis(np.array([1, 0, 1, 1], bool), 0, 4, xout)
    >>> print("".join('1' if xx else '0' for xx in xout))
    1001
    >>> __w_model_epistasis(np.array([1, 1, 0, 1], bool), 0, 4, xout)
    >>> print("".join('1' if xx else '0' for xx in xout))
    0101
    >>> __w_model_epistasis(np.array([1, 1, 1, 0], bool), 0, 4, xout)
    >>> print("".join('1' if xx else '0' for xx in xout))
    0011
    >>> __w_model_epistasis(np.array([0, 0, 1, 1], bool), 0, 4, xout)
    >>> print("".join('1' if xx else '0' for xx in xout))
    0110
    >>> __w_model_epistasis(np.array([0, 1, 0, 1], bool), 0, 4, xout)
    >>> print("".join('1' if xx else '0' for xx in xout))
    1010
    >>> __w_model_epistasis(np.array([0, 1, 1, 0], bool), 0, 4, xout)
    >>> print("".join('1' if xx else '0' for xx in xout))
    1100
    >>> __w_model_epistasis(np.array([1, 0, 0, 1], bool), 0, 4, xout)
    >>> print("".join('1' if xx else '0' for xx in xout))
    0010
    >>> __w_model_epistasis(np.array([1, 0, 1, 0], bool), 0, 4, xout)
    >>> print("".join('1' if xx else '0' for xx in xout))
    0100
    >>> __w_model_epistasis(np.array([1, 1, 0, 0], bool), 0, 4, xout)
    >>> print("".join('1' if xx else '0' for xx in xout))
    1000
    """
    end: Final[int] = (start + nu) - 1
    flip: Final[bool] = x_in[start]
    i: int = end
    skip: int = start
    while i >= start:
        result: bool = flip
        j: int = end
        while j > start:
            if j != skip:
                result = result ^ x_in[j]
            j = j - 1
        x_out[i] = result
        skip = skip - 1
        if skip < start:
            skip = end
        i = i - 1


@numba.njit(nogil=True, cache=True, inline="always")
def w_model_epistasis(x_in: np.ndarray, nu: int, x_out: np.ndarray) -> None:
    """
    Perform the epistasis transformation.

    :param x_in: the input string
    :param nu: the epistasis parameter
    :param x_out: the output string
    """
    total_len: Final[int] = len(x_in)
    end: Final[int] = total_len - nu
    i: int = 0
    while i <= end:
        __w_model_epistasis(x_in, i, nu, x_out)
        i = i + nu
    if i < total_len:
        __w_model_epistasis(x_in, i, total_len - i, x_out)


@numba.njit(nogil=True, cache=True, inline="always")
def w_model_max_gamma(nopt: int) -> int:
    """
    Compute the maximum gamma value for a given length `nopt`.

    :param nopt: the length of the optimal bit string
    :return: the maximum gamma value
    """
    return int((nopt * (nopt - 1)) >> 1)


@numba.njit(nogil=False, cache=True, inline="always")
def w_model_create_ruggedness_permutation(
        gamma: int, nopt: int, r: array) -> None:
    """
    Create the raw ruggedness array.

    This function creates the ruggedness permutation where groups of rugged
    transformations alternate with deceptive permutations for increasing
    gamma.

    :param gamma: the parameter for the ruggedness
    :param nopt: the length of the optimal string
    :param r: the array to receive the permutation

    >>> r = array('i', range(11))
    >>> w_model_create_ruggedness_permutation(0, 10, r)
    >>> "".join(str(xx) for xx in r)
    '012345678910'
    >>> r = array('i', range(7))
    >>> w_model_create_ruggedness_permutation(9, 6, r)
    >>> r[3]
    3
    >>> r[6]
    5
    """
    max_gamma: Final[int] = w_model_max_gamma(nopt)
    start: int = 0 if gamma <= 0 else (
        nopt - 1 - int(0.5 + sqrt(0.25 + ((max_gamma - gamma) << 1))))
    k: int = 0
    j: int = 1
    while j <= start:
        if (j & 1) != 0:
            r[j] = nopt - k
        else:
            k = k + 1
            r[j] = k
        j = j + 1

    while j <= nopt:
        k = k + 1
        r[j] = (nopt - k) if ((start & 1) != 0) else k
        j = j + 1

    upper: Final[int] = ((gamma - max_gamma) + (
        ((nopt - start - 1) * (nopt - start)) >> 1))
    j = j - 1
    i: int = 1
    while i <= upper:
        j = j - 1
        r[j], r[nopt] = r[nopt], r[j]
        i = i + 1


@numba.njit(nogil=True, cache=True, inline="always")
def w_model_ruggedness_translate(gamma: int, nopt: int) -> int:
    """
    Transform the gamma values such that deceptive follows rugged.

    If the ruggedness permutations are created with raw gamma values, then
    rugged and deceptive permutations will alternate with rising gamma.
    With this function here, the gamma parameters are translated such that
    first all the rugged permutations come (with rising degree of ruggedness)
    and then the deceptive ones come.

    :param gamma: the parameter for the ruggedness
    :param nopt: the length of the optimal string
    :returns: the translated gamma values

    >>> w_model_ruggedness_translate(12, 6)
    9
    >>> w_model_ruggedness_translate(34, 25)
    57
    >>> w_model_ruggedness_translate(0, 5)
    0
    >>> w_model_ruggedness_translate(1, 5)
    1
    >>> w_model_ruggedness_translate(2, 5)
    2
    >>> w_model_ruggedness_translate(3, 5)
    3
    >>> w_model_ruggedness_translate(4, 5)
    4
    >>> w_model_ruggedness_translate(5, 5)
    8
    >>> w_model_ruggedness_translate(6, 5)
    9
    >>> w_model_ruggedness_translate(7, 5)
    10
    >>> w_model_ruggedness_translate(8, 5)
    7
    >>> w_model_ruggedness_translate(9, 5)
    6
    >>> w_model_ruggedness_translate(10, 5)
    5
    """
    if gamma <= 0:
        return 0

    last_upper: int = (nopt >> 1) * ((nopt + 1) >> 1)
    if gamma <= last_upper:
        j = int(((nopt + 2) * 0.5) - sqrt(
            (((nopt * nopt) * 0.25) + 1) - gamma))
        k = ((gamma - ((nopt + 2) * j)) + (j * j) + nopt)
        return int((k + 1 + ((((nopt + 2) * j) - (j * j) - nopt - 1) << 1))
                   - (j - 1))

    j = int((((nopt % 2) + 1) * 0.5) + sqrt(
        (((1 - (nopt % 2)) * 0.25) + gamma) - 1 - last_upper))
    k = gamma - (((j - (nopt % 2)) * (j - 1)) + 1 + last_upper)
    return int(w_model_max_gamma(nopt) - k - ((2 * j * j) - j)
               - ((nopt % 2) * ((-2 * j) + 1)))


@numba.njit(nogil=True, cache=True, inline="always")
def _w_model(x_in: np.ndarray, m: int, m_out: np.ndarray | None,
             nu: int, nu_out: np.ndarray | None, r: array | None) -> int:
    """
    Compute the value of the multi-stage W-model.

    :param x_in: the input array
    :param m: the neutrality level
    :param m_out: the temporary variable for the de-neutralized `x_in`
    :param nu: the epistasis level
    :param nu_out: the temporary variable for the epistasis transformed string
    :param r: the ruggedness transform, if any
    :returns: the objective value
    """
    neutral: np.ndarray
    if m_out is None:
        neutral = x_in
    else:
        neutral = m_out
        w_model_neutrality(x_in, m, neutral)
    epi: np.ndarray
    if nu_out is None:
        epi = neutral
    else:
        epi = nu_out
        w_model_epistasis(neutral, nu, epi)
    f = w_model_f(epi)
    if r is None:
        return int(f)
    return r[f]


class WModel(BitStringProblem):
    """The tunable W-Model benchmark problem."""

    def __init__(self, nopt: int, m: int = 1, nu: int = 2,
                 gamma: int = 0, name: str | None = None) -> None:
        """
        Initialize an W-Model instance.

        :param nopt: the length of the optimal string
        :param m: the neutrality parameter
        :param nu: the epistasis parameter
        :param gamma: the ruggedness parameter
        :param name: the (optional) special name of this instance
        """
        super().__init__(nopt * m)
        #: the length of the optimal string
        self.nopt: Final[int] = check_int_range(nopt, "nopt", 2)
        #: the neutrality parameter
        self.m: Final[int] = check_int_range(m, "m", 1)
        #: the internal buffer for de-neutralization
        self.__m_out: Final[np.ndarray | None] = None if m < 2 \
            else np.empty(nopt, DEFAULT_BOOL)
        #: the epistasis parameter
        self.nu: Final[int] = check_int_range(nu, "nu", 2, nopt)
        #: the normalized epistasis parameter
        self.nu1: Final[float] = (nu - 2) / (nopt - 2)
        #: the internal buffer for de-epistazation
        self.__nu_out: Final[np.ndarray | None] = None if nu <= 2 \
            else np.empty(nopt, DEFAULT_BOOL)
        max_gamma: Final[int] = w_model_max_gamma(nopt)
        #: the ruggedness parameter
        self.gamma: Final[int] = check_int_range(gamma, "gamma", 0, max_gamma)
        #: the normalized ruggedness parameter
        self.gamma1: Final[float] = gamma / max_gamma
        #: the translated gamma parameter
        self.gamma_prime: Final[int] = w_model_ruggedness_translate(
            gamma, nopt)
        r: array | None = None  # the ruggedness table
        if gamma > 0:
            r = array("i", range(nopt + 1))
            w_model_create_ruggedness_permutation(self.gamma_prime, nopt, r)
        #: the ruggedness table
        self.__r: Final[array | None] = r

        if name is None:
            name = f"wmodel_{self.nopt}_{self.m}_{self.nu}_{self.gamma}"
        else:
            if not isinstance(name, str):
                raise type_error(name, "name", str)
            if (len(name) <= 0) or (sanitize_name(name) != name):
                raise ValueError(f"invalid name: {name!r}")
        #: the name of this w-model instance
        self.name: Final[str] = name

    def upper_bound(self) -> int:
        """
        Get the upper bound of the bit string based problem.

        :return: the length of the bit string

        >>> print(WModel(7, 6, 4, 4).upper_bound())
        7
        """
        return self.nopt

    def evaluate(self, x: np.ndarray) -> int:
        """
        Evaluate a solution to the W-Model.

        :param x: the bit string to evaluate
        :returns: the value of the W-Model for the string
        """
        return _w_model(x, self.m, self.__m_out, self.nu,
                        self.__nu_out, self.__r)

    def __str__(self):
        """
        Get the name of the W-Model instance.

        :returns: the name of the W-Model instance.
        """
        return self.name

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of this component as key-value pairs.

        :param logger: the logger for the parameters

        >>> from moptipy.utils.logger import InMemoryLogger
        >>> with InMemoryLogger() as l:
        ...     with l.key_values("C") as kv:
        ...         WModel(6, 2, 4, 7).log_parameters_to(kv)
        ...     text = l.get_log()
        >>> text[1]
        'name: wmodel_6_2_4_7'
        >>> text[3]
        'lowerBound: 0'
        >>> text[4]
        'upperBound: 6'
        >>> text[5]
        'n: 12'
        >>> text[6]
        'nopt: 6'
        >>> text[7]
        'm: 2'
        >>> text[8]
        'nu: 4'
        >>> text[9]
        'nu1: 0.5'
        >>> text[11]
        'gamma: 7'
        >>> text[12]
        'gamma1: 0.4666666666666667'
        >>> text[14]
        'gammaPrime: 11'
        >>> len(text)
        16
        """
        super().log_parameters_to(logger)
        logger.key_value("nopt", self.nopt)
        logger.key_value("m", self.m)
        logger.key_value("nu", self.nu)
        logger.key_value("nu1", self.nu1)
        logger.key_value("gamma", self.gamma)
        logger.key_value("gamma1", self.gamma1)
        logger.key_value("gammaPrime", self.gamma_prime)

    @staticmethod
    def default_instances() -> Iterable[Callable[[], "WModel"]]:
        """
        Get the 19 default instances of the W-Model.

        :returns: an `Iterable` that can provide callables constructing the
            19 default instances of the W-Model

        >>> len(list(WModel.default_instances()))
        19
        """
        return (cast(Callable[[], "WModel"],
                     lambda a=iid, b=z[0], c=z[1], d=z[2], g=z[3]:
                     WModel(b, c, d, g, f"wmodel{a + 1}"))
                for iid, z in enumerate([
                    (10, 2, 6, 10), (10, 2, 6, 18), (16, 1, 5, 72),
                    (16, 3, 9, 72), (25, 1, 23, 90), (32, 1, 2, 397),
                    (32, 4, 11, 0), (32, 4, 14, 0), (32, 4, 8, 128),
                    (50, 1, 36, 245), (50, 2, 21, 256), (50, 3, 16, 613),
                    (64, 2, 32, 256), (64, 3, 21, 16), (64, 3, 21, 256),
                    (64, 3, 21, 403), (64, 4, 52, 2), (75, 1, 60, 16),
                    (75, 2, 32, 4)]))



"""
Computations regarding the size of the solution spaces in a JSSP.

We represent solutions for a Job Shop Scheduling Problems (JSSPs) as Gantt
diagrams.
Assume that we look at JSSPs with n jobs and m machines.
The number of possible Gantt charts (without useless delays) is
(jobs!)**machines.

However, not all of them are necessarily feasible.
If all jobs pass through all machines in the same order, then all of the
possible Gantt charts are also feasible.
However, if, say job 0 first goes to machine 0 and then to machine 1 and
job 1 first goes to machine 1 and then to machine 0, there are possible
Gantt charts with deadlocks: If we put the second operation of job 0 to
be the first operation to be done by machine 1 and put the second operation
of job 1 to be the first operation to be done by machine 0, we end up with
an infeasible Gantt chart, i.e., one that cannot be executed.
Thus, the question arises: "For a given number n of jobs and m of machines,
what is the instance with the fewest feasible Gantt charts?"

Well, I am sure that there are clever algorithms to compute this. Maybe we
can even ave an elegant combinatorial formula.
But I do not want to spend much time on this subject (and maybe would not
be able to figure it out even if I wanted to...).
So we try to find this information using a somewhat brute force approach:
By enumerating instances and, for each instance, the Gantt charts, and
count how many of them are feasible.
"""
import sys
from math import factorial, log10
from typing import Iterable

import numba  # type: ignore
import numpy as np

from moptipy.examples.jssp import experiment
from moptipy.examples.jssp.gantt_space import gantt_space_size
from moptipy.examples.jssp.instance import Instance
from moptipy.utils.lang import Lang
from moptipy.utils.path import Path
from moptipy.utils.types import check_int_range


def permutations_with_repetitions_space_size(n: int, m: int) -> int:
    """
    Compute the number of n-permutations with m repetitions.

    :param n: the number of different values
    :param m: the number of repetitions
    :returns: the space size
    :rtype: int
    """
    return factorial(n * m) // (factorial(m) ** n)


#: the pre-computed values
__PRE_COMPUTED: tuple[tuple[int, int, int,
                            tuple[tuple[int, ...], ...]], ...] = (
    (3, 2, 22, ((0, 1), (0, 1), (1, 0))),
    (3, 3, 63, ((0, 1, 2), (1, 0, 2), (2, 0, 1))),
    (3, 4, 147, ((0, 1, 2, 3), (1, 0, 3, 2), (2, 3, 0, 1))),
    (3, 5, 317, ((0, 1, 2, 3, 4), (2, 1, 0, 4, 3), (3, 4, 0, 2, 1))),
    (4, 2, 244, ((0, 1), (0, 1), (1, 0), (1, 0))),
    (4, 3, 1630, ((0, 1, 2), (1, 0, 2), (2, 0, 1), (2, 1, 0))),
    (4, 4, 7451, ((0, 1, 2, 3), (1, 0, 3, 2), (2, 3, 1, 0), (3, 2, 0, 1))),
    (5, 2, 4548, ((0, 1), (0, 1), (0, 1), (1, 0), (1, 0))),
    (5, 3, 91461, ((0, 1, 2), (0, 2, 1), (1, 0, 2), (1, 2, 0), (2, 0, 1))),
    (6, 2, 108828, ((0, 1), (0, 1), (0, 1), (1, 0), (1, 0), (1, 0))),
    (7, 2, 3771792, ((0, 1), (0, 1), (0, 1), (0, 1), (1, 0), (1, 0), (1, 0))),
    (8, 2, 156073536,
     ((0, 1), (0, 1), (0, 1), (0, 1), (1, 0), (1, 0), (1, 0), (1, 0))),
)


def gantt_min_feasible(jobs: int, machines: int) \
        -> tuple[int, tuple[tuple[int, ...], ...]]:
    """
    Find the minimum number of feasible gantt charts.

    :param jobs: the number of jobs
    :param machines: the number of machines
    :return: the minimum number of feasible solutions for any instance
        of the given configuration and one example of such an instance
    """
    check_int_range(jobs, "jobs", 1, 127)
    check_int_range(machines, "machines", 1, 127)

    if machines <= 1:
        return factorial(jobs), (tuple([0] * jobs), )
    if jobs <= 1:
        return 1, (tuple(range(machines)), )
    if jobs <= 2:
        return machines + 1, (
            (*list(range(machines - 2, -1, -1)), machines - 1),
            (machines - 1, *list(range(machines - 1))))

    for tup in __PRE_COMPUTED:
        if (tup[0] == jobs) and (tup[1] == machines):
            return tup[2], tup[3]

    if machines <= 2:  # if there are two machines, we know the shape
        lst = [[0, 1]] * (jobs - (jobs // 2))
        lst.extend([[1, 0]] * (jobs // 2))
        dest = np.array(lst, dtype=np.uint8)
        res = __enumerate_feasible_for(jobs, machines, dest)
    else:  # more than two machines: need to enumerate
        dest = np.ndarray(shape=(jobs, machines), dtype=np.uint8)
        res = int(__find_min_feasible(np.int64(jobs),
                                      np.int64(machines), dest))

    # turn the result into a tuple
    arr = tuple(sorted([tuple([int(dest[i, j]) for j in range(machines)])
                        for i in range(jobs)]))
    return res, arr


@numba.njit
def __copy(dest: np.ndarray, source: np.ndarray, n: np.int64) -> None:
    """
    Copy an array.

    :param dest: the destination
    :param source: the source
    :param n: the number of elements to copy
    """
    for a in range(n):
        dest[a] = source[a]


@numba.njit
def __copy_instance(dest: np.ndarray, source: np.ndarray,
                    jobs: np.int64, machines: np.int64) -> None:
    """
    Copy an instance.

    :param dest: the destination
    :param source: the source
    :param jobs: the number of jobs
    :param machines: the machines
    """
    for a in range(jobs):
        __copy(dest[a], source[a], machines)


@numba.njit
def __find_min_feasible(jobs: np.int64, machines: np.int64,
                        dest: np.ndarray) -> np.int64:
    """
    Find the minimum number of feasible gantt charts.

    :param jobs: the number of jobs
    :param machines: the number of machines
    :param dest: the destination array
    :return: the minimum number of feasible solutions for any instance
        of the given configuration
    """
    instance = np.empty(shape=(jobs, machines), dtype=np.uint8)
    gantt = np.empty(shape=(machines, jobs), dtype=np.uint8)
    gantt_index = np.zeros(machines, dtype=np.int64)
    inst_index = np.zeros(jobs, dtype=np.int64)
    job_state = np.zeros(jobs, dtype=np.int64)
    gantt_state = np.zeros(machines, dtype=np.int64)
    upper_bound: np.int64 = np.int64(9223372036854775807)

    for i in range(jobs):
        __first_perm(instance[i], inst_index, i, machines)

    while True:
        if __check_sorted(instance, jobs, machines):
            upper_bound = __enumerate_feasible(
                instance, gantt, job_state, gantt_state, jobs, machines,
                upper_bound, gantt_index, dest)

        k = jobs - 1
        while True:
            if __next_perm(instance[k], inst_index, k, machines):
                break
            k = k - 1
            if k < 0:
                return upper_bound
        for j in range(k + 1, jobs):
            __first_perm(instance[j], inst_index, j, machines)


@numba.njit(nogil=True)
def __check_sorted(instance: np.ndarray,
                   jobs: np.int64,
                   machines: np.int64) -> bool:
    """
    Check if the instance is such that all jobs are sorted.

    :param instance: the instance
    :param jobs: the number of jobs
    :param machines: the number of machines
    :return: `True` if the instance is sorted, `False` otherwise
    """
    i: np.int64 = jobs - 1
    arr1: np.ndarray = instance[i]
    while i > 0:
        arr2 = arr1
        i = i - 1
        arr1 = instance[i]
        for j in range(machines):
            if arr1[j] > arr2[j]:
                return True
            if arr1[j] < arr2[j]:
                return False
    return True


@numba.njit
def __is_feasible(instance: np.ndarray,
                  gantt: np.ndarray,
                  job_state: np.ndarray,
                  gantt_state: np.ndarray,
                  jobs: np.int64,
                  machines: np.int64,
                  row: np.int64) -> bool:
    """
    Check if a Gantt diagram populated until a given row is feasible.

    :param instance: the JSSP instance, size jobs*machines
    :param gantt: the gantt chart, size machines*jobs
    :param job_state: the job state, of length jobs
    :param gantt_state: the machine state, of length machines
    :param jobs: the number of jobs
    :param machines: the number of machines
    :param row: the number of valid rows of the Gantt chart
    :return: `True` if the chart is feasible so far, `False` otherwise
    """
    if row <= 1:
        return True
    job_state.fill(0)  # all jobs start at the 0'th operations
    gantt_state.fill(0)  # all machines start at op 0
    found: bool = True
    needed_jobs = jobs  # the number of required jobs

    while found:
        found = False
        for job in range(jobs):  # check all jobs
            while True:  # we process each job as long as we can
                js = job_state[job]  # which operation is required?
                if js >= machines:
                    break  # the job is already finished
                nm = instance[job, js]  # the machine that the operation needs
                if nm >= row:  # ok, this machine is outside of the chart
                    js += 1  # so we assume it's ok
                    job_state[job] = js  # and step forward the state
                    found = True  # we did something!
                    if js >= machines:  # oh, we finished the job?
                        needed_jobs -= 1  # one less jobs to do
                        if needed_jobs <= 0:  # no more jobs to do?
                            return True  # the chart is feasible!
                        break  # quit handling this job, as its finished
                    continue  # move to next operation, as job is not finished
                ms = gantt_state[nm]  # get next item in gantt chart
                mj = gantt[nm, ms]  # next job on this machine?
                if mj == job:  # great, this job!
                    gantt_state[nm] = ms + 1  # advance gantt state
                    js += 1  # so we can perform the operation
                    job_state[job] = js  # and step forward the state
                    found = True  # we did something!
                    if js >= machines:  # oh, we finished the job?
                        needed_jobs -= 1  # one less jobs to do
                        if needed_jobs <= 0:  # no more jobs to do?
                            return True  # the chart is feasible!
                        break  # quit handling this job, as its finished
                    continue  # move to next operation, as job is not finished
                break  # no, we cannot handle this job now

    return False  # we ended one round without being able to proceed


@numba.njit
def __first_perm(arr: np.ndarray,
                 index: np.ndarray,
                 pi: np.int64,
                 n: np.int64) -> None:
    """
    Create the first permutation for a given array of values.

    :param arr: the array to permute over
    :param index: the array with the index
    :param pi: the index of the index to use in pi
    :param n: the length of arr
    """
    for i in range(n):
        arr[i] = i
    index[pi] = np.int64(0)


@numba.njit
def __next_perm(arr: np.ndarray,
                index: np.ndarray,
                pi: np.int64,
                n: np.int64) -> bool:
    """
    Get the next permutation for a given array of values.

    :param arr: the array to permute over
    :param index: the array with the index
    :param pi: the index of the index to use in pi
    :param n: the length of arr
    :returns: `True` if there is a next permutation, `False` if not
    """
    idx = index[pi]
    if idx >= n - 1:
        return False

    nidx = idx + 1

    if idx == 0:  # increase is at the very beginning
        arr[0], arr[1] = arr[1], arr[0]  # swap
        idx = nidx
        while True:  # update index to the next increase
            nidx = idx + 1
            if nidx >= n:
                break  # reached end
            if arr[idx] <= arr[nidx]:
                break  # found increase
            idx = nidx
    else:
        if arr[nidx] > arr[0]:  # value at arr[idx + 1] is greater than arr[0]
            # no need for binary search, just swap arr[idx + 1] and arr[0]
            arr[nidx], arr[0] = arr[0], arr[nidx]
        else:
            # binary search to find the greatest value which is less
            # than arr[idx + 1]
            start = np.int64(0)
            end = idx
            mid = (start + end) // 2
            t_value = arr[nidx]
            while not (arr[mid] < t_value < arr[mid - 1]):
                if arr[mid] < t_value:
                    end = mid - 1
                else:
                    start = mid + 1
                mid = (start + end) // 2
            arr[nidx], arr[mid] = arr[mid], arr[nidx]  # swap

        # invert 0 to increase
        for i in range((idx // 2) + 1):
            arr[i], arr[idx - i] = arr[idx - i], arr[i]
        idx = 0  # reset increase

    index[pi] = idx
    return True


@numba.njit
def __enumerate_feasible(instance: np.ndarray,
                         gantt: np.ndarray,
                         job_state: np.ndarray,
                         gantt_state: np.ndarray,
                         jobs: np.int64,
                         machines: np.int64,
                         upper_bound: np.int64,
                         index: np.ndarray,
                         dest: np.ndarray) -> np.int64:
    """
    Enumerate the feasible gantt charts for an instance.

    :param instance: the JSSP instance, size jobs*machines
    :param gantt: the gantt chart array, size machines*jobs
    :param job_state: the job state, of length jobs
    :param gantt_state: the machine state, of length machines
    :param jobs: the number of jobs
    :param machines: the number of machines
    :param upper_bound: the upper bound - we won't enumerate more
        charts than this.
    :param index: the index array
    :param dest: the destination array
    :returns: the number of enumerated feasible gantt charts
    """
    counter: np.int64 = np.int64(0)
    for z in range(machines):
        __first_perm(gantt[z], index, z, jobs)

    while True:
        if __is_feasible(instance, gantt, job_state, gantt_state,
                         jobs, machines, machines):
            counter += 1  # found another feasible gantt chart for instance
            if counter >= upper_bound:
                return counter  # if we have reached the minimum, we can stop

        i = machines - 1
        while True:
            if __next_perm(gantt[i], index, i, jobs):
                if i >= machines - 1:
                    break
                if __is_feasible(instance, gantt, job_state, gantt_state,
                                 jobs, machines, i + 1):
                    for j in range(i + 1, machines):
                        __first_perm(gantt[j], index, j, jobs)
                    break
            else:
                i -= 1
                if i < 0:
                    if instance is not dest:
                        __copy_instance(dest, instance, jobs, machines)
                    return counter  # we have enumerated all gantt charts


@numba.njit
def __enumerate_feasible_for(jobs: np.int64, machines: np.int64,
                             instance: np.ndarray) -> np.int64:
    """
    Find the minimum number of feasible gantt charts.

    :param jobs: the number of jobs
    :param machines: the number of machines
    :param instance: the provided instance array
    :return: the minimum number of feasible solutions for any instance
        of the given configuration
    """
    gantt = np.empty(shape=(machines, jobs), dtype=np.uint8)
    gantt_index = np.zeros(machines, dtype=np.int64)
    job_state = np.zeros(jobs, dtype=np.int64)
    gantt_state = np.zeros(machines, dtype=np.int64)
    upper_bound: np.int64 = np.int64(9223372036854775807)
    return __enumerate_feasible(instance, gantt, job_state,
                                gantt_state, jobs, machines,
                                upper_bound, gantt_index,
                                instance)


def __long_str(value: int) -> str:
    """
    Convert a value to a string.

    :param value: the value
    :returns: the string representation
    """
    if value < 0:
        return ""
    if value <= 1_000_000_000_000:
        return Lang.current().format_int(value)
    logg = log10(value)
    exp = int(logg)
    base = value / (10 ** exp)
    expf = Lang.current().format_int(exp)
    return f"$\\approx$&nbsp;{base:.3f}*10^{expf}^"


def make_gantt_space_size_table(
        dest: str = "solution_space_size.md",
        instances: Iterable[str] = tuple(list(  # noqa
            experiment.INSTANCES) + ["demo"])) -> Path:  # noqa
    """
    Print a table of solution space sizes.

    :param dest: the destination file
    :param instances: the instances to add
    :returns: the fully-qualified path to the generated file
    """
    file = Path.path(dest)
    text = [f'|{Lang.current()["name"]}|'
            + r"$\jsspJobs$|$\jsspMachines$|$\min(\#\text{"
            + f'{Lang.current()["feasible"]}'
            + r"})$|$\left|\solutionSpace\right|$|",
            r"|:--|--:|--:|--:|--:|"]

    inst_scales: list[tuple[int, int, int, int, str]] = []

    # enumerate the pre-defined instances
    for inst in set(instances):
        instance = Instance.from_resource(inst)
        min_size = -1
        for tup in __PRE_COMPUTED:
            if (tup[0] == instance.jobs) and (tup[1] == instance.machines):
                min_size = tup[2]
                break
        if (min_size < 0) and ((instance.jobs <= 2)
                               or (instance.machines <= 2)):
            min_size = gantt_min_feasible(
                instance.jobs, instance.machines)[0]
        inst_scales.append(
            (instance.jobs, instance.machines,
             gantt_space_size(instance.jobs, instance.machines),
             min_size, f"`{instance.name}`"))
        del instance

    # enumerate some default values
    for jobs in range(2, 6):
        for machines in range(2, 6):
            found: bool = False  # skip over already added scales
            for tupp in inst_scales:
                if (tupp[0] == jobs) and (tupp[1] == machines):
                    found = True
                    break
            if found:
                continue

            min_size = -1
            for tup in __PRE_COMPUTED:
                if (tup[0] == jobs) and (tup[1] == machines):
                    min_size = tup[2]
                    break
            if (min_size < 0) and ((jobs <= 2) or (machines <= 2)):
                min_size = gantt_min_feasible(jobs, machines)[0]
            name = "[@fig:jssp_feasible_gantt]" \
                if (jobs == 2) and (machines == 2) else ""
            inst_scales.append(
                (jobs, machines, gantt_space_size(jobs, machines),
                 min_size, name))

    inst_scales.sort()
    for i, ua in enumerate(inst_scales):
        a = ua
        for j in range(i, len(inst_scales)):
            b = inst_scales[j]
            if (a[-1] and b[-1]) and (a[-3] > b[-3]):
                inst_scales[i] = b
                inst_scales[j] = a
                a = b

    for scale in inst_scales:
        text.append(f"|{scale[4]}|{scale[0]}|{scale[1]}|"
                    f"{__long_str(scale[3])}|{__long_str(scale[2])}|")

    file.write_all(text)
    file.enforce_file()
    return file


def make_search_space_size_table(
        dest: str = "solution_space_size.md",
        instances: Iterable[str] = tuple(list(  # noqa
            experiment.INSTANCES) + ["demo"])) -> Path:  # noqa
    """
    Print a table of search space sizes.

    :param dest: the destination file
    :param instances: the instances to add
    :returns: the fully-qualified path to the generated file
    """
    file = Path.path(dest)
    text = [f'|{Lang.current()["name"]}|'
            + r"$\jsspJobs$|$\jsspMachines$|$\left|\solutionSpace\right|$|"
            + r"$\left|\searchSpace\right|$|",
            r"|:--|--:|--:|--:|--:|"]
    inst_scales: list[tuple[int, int, int, int, str]] = []

    # enumerate the pre-defined instances
    for inst in set(instances):
        instance = Instance.from_resource(inst)
        inst_scales.append(
            (instance.jobs, instance.machines,
             gantt_space_size(instance.jobs, instance.machines),
             permutations_with_repetitions_space_size(
                 instance.jobs, instance.machines),
             f"`{instance.name}`"))
        del instance

    # enumerate some default values
    for jobs in range(3, 6):
        for machines in range(2, 6):
            found: bool = False  # skip over already added scales
            for tupp in inst_scales:
                if (tupp[0] == jobs) and (tupp[1] == machines):
                    found = True
                    break
            if found:
                continue
            inst_scales.append(
                (jobs, machines, gantt_space_size(jobs, machines),
                 permutations_with_repetitions_space_size(
                     jobs, machines), ""))

    inst_scales.sort()
    for i, ua in enumerate(inst_scales):
        a = ua
        for j in range(i, len(inst_scales)):
            b = inst_scales[j]
            if (a[-1] and b[-1]) and (a[-2] > b[-2]):
                inst_scales[i] = b
                inst_scales[j] = a
                a = b
    for scale in inst_scales:
        text.append(f"|{scale[4]}|{scale[0]}|{scale[1]}|"
                    f"{__long_str(scale[2])}|{__long_str(scale[3])}|")

    file.write_all(text)
    file.enforce_file()
    return file


# create the tables if this is the main script
if __name__ == "__main__":
    dest_dir = Path.path(sys.argv[1])
    dest_dir.ensure_dir_exists()
    for lang in Lang.all_langs():
        lang.set_current()
        make_gantt_space_size_table(
            dest_dir.resolve_inside(
                lang.filename("solution_space_size") + ".md"))
        make_search_space_size_table(
            dest_dir.resolve_inside(
                lang.filename("search_space_size") + ".md"))



"""
Plot a set of :class:`~moptipy.evaluation.ert.Ert` objects into one figure.

The (empirically estimated) Expected Running Time (ERT, see
:mod:`~moptipy.evaluation.ert`) is a function that tries to give an estimate
how long a given algorithm setup will need (y-axis) to achieve given solution
qualities (x-axis). It uses a set of runs of the algorithm on the problem to
make this estimate under the assumption of independent restarts.

1. Kenneth V. Price. Differential Evolution vs. The Functions of the 2nd ICEO.
   In Russ Eberhart, Peter Angeline, Thomas Back, Zbigniew Michalewicz, and
   Xin Yao, editors, *IEEE International Conference on Evolutionary
   Computation,* April 13-16, 1997, Indianapolis, IN, USA, pages 153-157.
   IEEE Computational Intelligence Society. ISBN: 0-7803-3949-5.
   doi: https://doi.org/10.1109/ICEC.1997.592287
2. Nikolaus Hansen, Anne Auger, Steffen Finck, Raymond Ros. *Real-Parameter
   Black-Box Optimization Benchmarking 2010: Experimental Setup.*
   Research Report RR-7215, INRIA. 2010. inria-00462481.
   https://hal.inria.fr/inria-00462481/document/
"""
from typing import Any, Callable, Final, Iterable, cast

import numpy as np
from matplotlib.artist import Artist  # type: ignore
from matplotlib.axes import Axes  # type: ignore
from matplotlib.figure import Figure  # type: ignore

import moptipy.utils.plot_defaults as pd
import moptipy.utils.plot_utils as pu
from moptipy.evaluation.axis_ranger import AxisRanger
from moptipy.evaluation.base import get_algorithm, get_instance, sort_key
from moptipy.evaluation.ert import Ert
from moptipy.evaluation.styler import Styler
from moptipy.utils.lang import Lang
from moptipy.utils.types import type_error


def plot_ert(erts: Iterable[Ert],
             figure: Axes | Figure,
             x_axis: AxisRanger | Callable[[str], AxisRanger]
             = AxisRanger.for_axis,
             y_axis: AxisRanger | Callable[[str], AxisRanger]
             = AxisRanger.for_axis,
             legend: bool = True,
             distinct_colors_func: Callable[[int], Any] =
             pd.distinct_colors,
             distinct_line_dashes_func: Callable[[int], Any] =
             pd.distinct_line_dashes,
             importance_to_line_width_func: Callable[[int], float] =
             pd.importance_to_line_width,
             importance_to_alpha_func: Callable[[int], float] =
             pd.importance_to_alpha,
             importance_to_font_size_func: Callable[[int], float] =
             pd.importance_to_font_size,
             x_grid: bool = True,
             y_grid: bool = True,
             x_label: None | str | Callable[[str], str] = Lang.translate,
             x_label_inside: bool = True,
             y_label: None | str | Callable[[str], str] =
             Lang.translate_func("ERT"),
             y_label_inside: bool = True,
             instance_sort_key: Callable[[str], Any] = lambda x: x,
             algorithm_sort_key: Callable[[str], Any] = lambda x: x,
             instance_namer: Callable[[str], str] = lambda x: x,
             algorithm_namer: Callable[[str], str] = lambda x: x,
             instance_priority: float = 0.666,
             algorithm_priority: float = 0.333) -> Axes:
    """
    Plot a set of Ert functions into one chart.

    :param erts: the iterable of Ert functions
    :param figure: the figure to plot in
    :param x_axis: the x_axis ranger
    :param y_axis: the y_axis ranger
    :param legend: should we plot the legend?
    :param distinct_colors_func: the function returning the palette
    :param distinct_line_dashes_func: the function returning the line styles
    :param importance_to_line_width_func: the function converting importance
        values to line widths
    :param importance_to_alpha_func: the function converting importance
        values to alphas
    :param importance_to_font_size_func: the function converting importance
        values to font sizes
    :param x_grid: should we have a grid along the x-axis?
    :param y_grid: should we have a grid along the y-axis?
    :param x_label: a callable returning the label for the x-axis, a label
        string, or `None` if no label should be put
    :param x_label_inside: put the x-axis label inside the plot (so that
        it does not consume additional vertical space)
    :param y_label: a callable returning the label for the y-axis, a label
        string, or `None` if no label should be put
    :param y_label_inside: put the y-axis label inside the plot (so that
        it does not consume additional horizontal space)
    :param instance_sort_key: the sort key function for instances
    :param algorithm_sort_key: the sort key function for algorithms
    :param instance_namer: the name function for instances receives an
        instance ID and returns an instance name; default=identity function
    :param algorithm_namer: the name function for algorithms receives an
        algorithm ID and returns an instance name; default=identity function
    :param instance_priority: the style priority for instances
    :param algorithm_priority: the style priority for algorithms
    :returns: the axes object to allow you to add further plot elements
    """
    # Before doing anything, let's do some type checking on the parameters.
    # I want to ensure that this function is called correctly before we begin
    # to actually process the data. It is better to fail early than to deliver
    # some incorrect results.
    if not isinstance(erts, Iterable):
        raise type_error(erts, "erts", Iterable)
    if not isinstance(figure, Axes | Figure):
        raise type_error(figure, "figure", (Axes, Figure))
    if not isinstance(legend, bool):
        raise type_error(legend, "legend", bool)
    if not callable(distinct_colors_func):
        raise type_error(
            distinct_colors_func, "distinct_colors_func", call=True)
    if not callable(distinct_line_dashes_func):
        raise type_error(
            distinct_line_dashes_func, "distinct_line_dashes_func", call=True)
    if not callable(distinct_line_dashes_func):
        raise type_error(importance_to_line_width_func,
                         "importance_to_line_width_func", call=True)
    if not callable(importance_to_alpha_func):
        raise type_error(
            importance_to_alpha_func, "importance_to_alpha_func", call=True)
    if not callable(importance_to_font_size_func):
        raise type_error(importance_to_font_size_func,
                         "importance_to_font_size_func", call=True)
    if not isinstance(x_grid, bool):
        raise type_error(x_grid, "x_grid", bool)
    if not isinstance(y_grid, bool):
        raise type_error(y_grid, "y_grid", bool)
    if not ((x_label is None) or callable(x_label)
            or isinstance(x_label, str)):
        raise type_error(x_label, "x_label", (str, None), call=True)
    if not isinstance(x_label_inside, bool):
        raise type_error(x_label_inside, "x_label_inside", bool)
    if not ((y_label is None) or callable(y_label)
            or isinstance(y_label, str)):
        raise type_error(y_label, "y_label", (str, None), call=True)
    if not isinstance(y_label_inside, bool):
        raise type_error(y_label_inside, "y_label_inside", bool)
    if not callable(instance_sort_key):
        raise type_error(instance_sort_key, "instance_sort_key", call=True)
    if not callable(algorithm_sort_key):
        raise type_error(algorithm_sort_key, "algorithm_sort_key", call=True)
    if not isinstance(instance_priority, float):
        raise type_error(instance_priority, "instance_priority", float)
    if not isinstance(algorithm_priority, float):
        raise type_error(algorithm_priority, "algorithm_priority", float)
    if not callable(instance_namer):
        raise type_error(instance_namer, "instance_namer", call=True)
    if not callable(algorithm_namer):
        raise type_error(algorithm_namer, "algorithm_namer", call=True)

    # First, we try to find groups of data to plot together in the same
    # color/style.
    instances: Final[Styler] = Styler(
        key_func=get_instance,
        namer=instance_namer,
        none_name=Lang.translate("all_insts"),
        priority=instance_priority,
        name_sort_function=instance_sort_key)
    algorithms: Final[Styler] = Styler(
        key_func=get_algorithm,
        namer=algorithm_namer,
        none_name=Lang.translate("all_algos"),
        priority=algorithm_priority,
        name_sort_function=algorithm_sort_key)
    x_dim: str | None = None
    y_dim: str | None = None
    source: list[Ert] = cast(list[Ert], erts) if isinstance(erts, list) \
        else list(erts)
    del erts

    # First pass: find out the instances and algorithms
    for ert in source:
        if not isinstance(ert, Ert):
            raise type_error(ert, "ert data source", Ert)
        instances.add(ert)
        algorithms.add(ert)

        # Validate that we have consistent time and objective units.
        if x_dim is None:
            x_dim = ert.f_name
        elif x_dim != ert.f_name:
            raise ValueError(
                f"F-units {x_dim} and {ert.f_name} do not fit!")

        if y_dim is None:
            y_dim = ert.time_unit
        elif y_dim != ert.time_unit:
            raise ValueError(
                f"Time units {y_dim} and {ert.time_unit} do not fit!")

    if (x_dim is None) or (y_dim is None) or (len(source) <= 0):
        raise ValueError("Illegal state?")

    # determine the style groups
    groups: list[Styler] = []
    instances.finalize()
    algorithms.finalize()

    sf: Callable[[Ert], Any] = sort_key
    if (instances.count > 1) and (algorithms.count == 1):
        def __x1(r: Ert, ssf=instance_sort_key) -> Any:
            return ssf(r.instance)
        sf = __x1
    elif (instances.count == 1) and (algorithms.count > 1):
        def __x2(r: Ert, ssf=algorithm_sort_key) -> Any:
            return ssf(r.algorithm)
        sf = __x2
    elif (instances.count > 1) and (algorithms.count > 1):
        def __x3(r: Ert, sas=algorithm_sort_key,
                 ias=instance_sort_key,
                 ag=algorithm_priority > instance_priority) \
                -> tuple[Any, Any]:
            k1 = ias(r.instance)
            k2 = sas(r.algorithm)
            return (k2, k1) if ag else (k1, k2)
        sf = __x3
    source.sort(key=sf)

    def __set_importance(st: Styler) -> None:
        none = 1
        not_none = 0
        none_lw = importance_to_line_width_func(none)
        not_none_lw = importance_to_line_width_func(not_none)
        st.set_line_width(lambda p: [none_lw if i <= 0 else not_none_lw
                                     for i in range(p)])
        none_a = importance_to_alpha_func(none)
        not_none_a = importance_to_alpha_func(not_none)
        st.set_line_alpha(lambda p: [none_a if i <= 0 else not_none_a
                                     for i in range(p)])

    if instances.count > 1:
        groups.append(instances)
    if algorithms.count > 1:
        groups.append(algorithms)

    if len(groups) > 0:
        groups.sort()
        groups[0].set_line_color(distinct_colors_func)

        if len(groups) > 1:
            groups[1].set_line_dash(distinct_line_dashes_func)

    # If we only have <= 2 groups, we can mark None and not-None values with
    # different importance.
    if instances.has_none and (instances.count > 1):
        __set_importance(instances)
    elif algorithms.has_none and (algorithms.count > 1):
        __set_importance(algorithms)

    # we will collect all lines to plot in plot_list
    plot_list: list[dict] = []

    # set up the axis rangers
    if callable(x_axis):
        x_axis = x_axis(x_dim)
    if not isinstance(x_axis, AxisRanger):
        raise type_error(x_axis, "x_axis", AxisRanger)

    if callable(y_axis):
        y_axis = y_axis(y_dim)
    if not isinstance(y_axis, AxisRanger):
        raise type_error(y_axis, "y_axis", AxisRanger)

    # first we collect all progress object
    for ert in source:
        style = pd.create_line_style()
        for g in groups:
            g.add_line_style(ert, style)
        x = ert.ert[:, 0]
        style["x"] = x
        x_axis.register_array(x)
        y = ert.ert[:, 1]
        y_axis.register_array(y)
        style["y"] = y
        plot_list.append(style)
    del source

    font_size_0: Final[float] = importance_to_font_size_func(0)

    # set up the graphics area
    axes: Final[Axes] = pu.get_axes(figure)
    axes.tick_params(axis="x", labelsize=font_size_0)
    axes.tick_params(axis="y", labelsize=font_size_0)

    # draw the grid
    if x_grid or y_grid:
        grid_lwd = importance_to_line_width_func(-1)
        if x_grid:
            axes.grid(axis="x", color=pd.GRID_COLOR, linewidth=grid_lwd)
        if y_grid:
            axes.grid(axis="y", color=pd.GRID_COLOR, linewidth=grid_lwd)

    max_y = y_axis.get_pinf_replacement()

    # plot the lines
    for line in plot_list:
        y = line["y"]
        if np.isposinf(y[0]):
            y = y.copy()
            y[0] = max_y
            line["y"] = y
        axes.step(where="post", **line)
    del plot_list

    x_axis.apply(axes, "x")
    y_axis.apply(axes, "y")

    if legend:
        handles: list[Artist] = []

        for g in groups:
            g.add_to_legend(handles.append)
            g.has_style = False

        if instances.has_style:
            instances.add_to_legend(handles.append)
        if algorithms.has_style:
            algorithms.add_to_legend(handles.append)

        axes.legend(loc="upper right",
                    handles=handles,
                    labelcolor=[art.color if hasattr(art, "color")
                                else pd.COLOR_BLACK for art in handles],
                    fontsize=font_size_0)

    pu.label_axes(axes=axes,
                  x_label=x_label(x_dim) if callable(x_label) else x_label,
                  x_label_inside=x_label_inside,
                  x_label_location=1,
                  y_label=y_label(y_dim) if callable(y_label) else y_label,
                  y_label_inside=y_label_inside,
                  y_label_location=0,
                  font_size=font_size_0)
    return axes



"""A binary operator copying each bit from either source string."""

import numba  # type: ignore
import numpy as np
from numpy.random import Generator

from moptipy.api.operators import Op2


@numba.njit(nogil=True, cache=True)
def __op2_uniform(r: np.ndarray, dest: np.ndarray, x0: np.ndarray,
                  x1: np.ndarray) -> None:
    """
    Perform the actual work of the uniform crossover.

    :param r: the array with random numbers in 0..1
    :param dest: the destination array
    :param x0: the first source array
    :param x1: the second source array
    """
    for i, v in enumerate(r):  # iterate over random numbers
        # copy from x0 with p=0.5 and from x1 with p=0.5
        dest[i] = x1[i] if v == 0 else x0[i]


def _op2_uniform(random: Generator, dest: np.ndarray, x0: np.ndarray,
                 x1: np.ndarray) -> None:
    """
    Perform the uniform operator as plain old function.

    :param random: the random number generator
    :param dest: the destination array to receive the new point
    :param x0: the first source array
    :param x1: the second source array
    """
    __op2_uniform(random.integers(low=2, high=None, size=len(dest)),
                  dest, x0, x1)


class Op2Uniform(Op2):
    """
    A binary search operation that copies each bit from either source.

    For each index `i` in the destination array `dest`, uniform
    crossover copies the value from the first source string `x0`with
    probability 0.5 and otherwise the value from the second source
    string `x1`. All bits that have the same value in `x0` and `x1`
    will retain this value in `dest`, all bits where `x0` and `x1`
    differ will effectively be randomized (be `0` with probability 0.5
    and `1` with probability 0.5).
    """

    def __init__(self):
        """Initialize the uniform crossover operator."""
        super().__init__()
        self.op2 = _op2_uniform  # type: ignore

    def __str__(self) -> str:
        """
        Get the name of this binary operator.

        :return: "uniform"
        """
        return "uniform"



"""Test stuff on permutations with repetitions."""

from typing import Any, Callable, Iterable, cast

import numpy as np
from numpy.random import Generator, default_rng

from moptipy.api.operators import Op0, Op1, Op1WithStepSize, Op2
from moptipy.spaces.permutations import Permutations
from moptipy.tests.op0 import validate_op0
from moptipy.tests.op1 import validate_op1
from moptipy.tests.op1_with_step_size import validate_op1_with_step_size
from moptipy.tests.op2 import validate_op2
from moptipy.utils.types import type_error


def permutations_for_tests(
        perm_filter: Callable[[Permutations], bool] | None = None) \
        -> Iterable[Permutations]:
    """
    Get a sequence of permutations for tests.

    :param perm_filter: an optional filter to sort out permutations we cannot
        use for testing
    :returns: the sequence of Permutations
    """
    r = default_rng()
    pwrs: list[Permutations] = [
        Permutations.standard(2),
        Permutations.standard(3),
        Permutations.standard(4),
        Permutations.standard(5),
        Permutations.standard(6),
        Permutations.standard(12),
        Permutations.standard(23),
        Permutations.with_repetitions(2, 2),
        Permutations.with_repetitions(2, 3),
        Permutations.with_repetitions(3, 2),
        Permutations.with_repetitions(3, 3),
        Permutations.with_repetitions(5, 5),
        Permutations.with_repetitions(int(r.integers(6, 10)),
                                      int(r.integers(2, 7))),
        Permutations.with_repetitions(int(r.integers(2, 5)),
                                      int(r.integers(6, 10))),
        Permutations.with_repetitions(int(r.integers(130, 500)),
                                      int(r.integers(2, 200))),
        Permutations([1, 1, 1, 1, 1, 5, 5, 3]),
        Permutations([0, 1, 1, 1, 1, 1])]
    if perm_filter is not None:
        if not callable(perm_filter):
            raise type_error(perm_filter, "perm_filter", None, call=True)
        pwrs = [p for p in pwrs if perm_filter(p)]
    r.shuffle(cast(list, pwrs))
    return pwrs


def make_permutation_valid(pwr: Permutations) -> \
        Callable[[Generator, np.ndarray], np.ndarray]:
    """
    Create a function that can make permutations with repetitions valid.

    :param pwr: the permutations
    :returns: the function
    """
    def __make_valid(prnd: Generator, x: np.ndarray, ppp=pwr) -> np.ndarray:
        np.copyto(x, ppp.blueprint)
        prnd.shuffle(x)
        return x
    return __make_valid


def validate_op0_on_1_permutations(
        op0: Op0 | Callable[[Permutations], Op0],
        search_space: Permutations,
        number_of_samples: int | None = None,
        min_unique_samples: int | Callable[[
            int, Permutations], int] | None = None) -> None:
    """
    Validate the nullary operator on one `Permutations` instance.

    :param op0: the operator or operator factory
    :param search_space: the search space
    :param number_of_samples: the optional number of samples
    :param min_unique_samples: the optional unique samples
    """
    args: dict[str, Any] = {
        "op0": op0(search_space) if callable(op0) else op0,
        "search_space": search_space,
        "make_search_space_element_valid":
            make_permutation_valid(search_space),
    }
    if number_of_samples is not None:
        args["number_of_samples"] = number_of_samples
    if min_unique_samples is not None:
        args["min_unique_samples"] = min_unique_samples
    validate_op0(**args)


def validate_op0_on_permutations(
        op0: Op0 | Callable[[Permutations], Op0],
        number_of_samples: int | None = None,
        min_unique_samples: int | Callable[[
            int, Permutations], int] | None = None,
        perm_filter: Callable[[Permutations], bool] | None = None) -> None:
    """
    Validate the nullary operator on several `Permutations` instances.

    :param op0: the operator or operator factory
    :param number_of_samples: the optional number of samples
    :param min_unique_samples: the optional unique samples
    :param perm_filter: an optional filter to sort out permutations we cannot
        use for testing
    """
    for pwr in permutations_for_tests(perm_filter):
        validate_op0_on_1_permutations(op0, pwr, number_of_samples,
                                       min_unique_samples)


def validate_op1_on_1_permutations(
        op1: Op1 | Callable[[Permutations], Op1],
        search_space: Permutations,
        number_of_samples: int | None = None,
        min_unique_samples: int | Callable[[
            int, Permutations], int] | None = None) -> None:
    """
    Validate the unary operator on one `Permutations` instance.

    :param op1: the operator or operator factory
    :param search_space: the search space
    :param number_of_samples: the optional number of samples
    :param min_unique_samples: the optional unique samples
    """
    args: dict[str, Any] = {
        "op1": op1(search_space) if callable(op1) else op1,
        "search_space": search_space,
        "make_search_space_element_valid":
            make_permutation_valid(search_space),
    }
    if number_of_samples is not None:
        args["number_of_samples"] = number_of_samples
    if min_unique_samples is not None:
        args["min_unique_samples"] = min_unique_samples
    validate_op1(**args)


def validate_op1_on_permutations(
        op1: Op1 | Callable[[Permutations], Op1],
        number_of_samples: int | None = None,
        min_unique_samples: int | Callable[[
            int, Permutations], int] | None = None,
        perm_filter: Callable[[Permutations], bool] | None = None) -> None:
    """
    Validate the unary operator on several `Permutations` instances.

    :param op1: the operator or operator factory
    :param number_of_samples: the optional number of samples
    :param min_unique_samples: the optional unique samples
    :param perm_filter: an optional filter to sort out permutations we cannot
        use for testing
    """
    for pwr in permutations_for_tests(perm_filter):
        validate_op1_on_1_permutations(op1, pwr, number_of_samples,
                                       min_unique_samples)


def validate_op1_with_step_size_on_1_permutations(
        op1: Op1WithStepSize | Callable[[Permutations], Op1WithStepSize],
        search_space: Permutations,
        number_of_samples: int | None = None,
        min_unique_samples: int | Callable[[
            int, Permutations], int] | None = None,
        step_sizes: Iterable[float] | Callable[
            [Permutations], Iterable[float]] = (),
        get_step_size: Callable[[
            Permutations, np.ndarray, np.ndarray,
        ], float | None] | None = None) -> None:
    """
    Validate the step-sized unary operator on one `Permutations` instance.

    :param op1: the operator or operator factory
    :param search_space: the search space
    :param number_of_samples: the optional number of samples
    :param min_unique_samples: the optional unique samples
    :param step_sizes: the step sizes to test
    :param get_step_size: try to get the step size from two space elements
    """
    args: dict[str, Any] = {
        "op1": op1(search_space) if callable(op1) else op1,
        "search_space": search_space,
        "make_search_space_element_valid":
            make_permutation_valid(search_space),
        "step_sizes": step_sizes(search_space) if callable(step_sizes)
        else step_sizes,
        "get_step_size": get_step_size,
    }
    if number_of_samples is not None:
        args["number_of_samples"] = number_of_samples
    if min_unique_samples is not None:
        args["min_unique_samples"] = min_unique_samples
    validate_op1_with_step_size(**args)


def validate_op1_with_step_size_on_permutations(
        op1: Op1WithStepSize | Callable[[Permutations], Op1WithStepSize],
        number_of_samples: int | None = None,
        min_unique_samples: int | Callable[[
            int, Permutations], int] | None = None,
        step_sizes: Iterable[float] | Callable[
            [Permutations], Iterable[float]] = (),
        get_step_size: Callable[[
            Permutations, np.ndarray, np.ndarray,
        ], float | None] | None = None,
        perm_filter: Callable[[Permutations], bool] | None = None) -> None:
    """
    Validate the unary operator on several `Permutations` instances.

    :param op1: the operator or operator factory
    :param number_of_samples: the optional number of samples
    :param min_unique_samples: the optional unique samples
    :param step_sizes: the step sizes to test
    :param get_step_size: try to get the step size from two space elements
    :param perm_filter: an optional filter to sort out permutations we cannot
        use for testing
    """
    for pwr in permutations_for_tests(perm_filter):
        validate_op1_with_step_size_on_1_permutations(
            op1, pwr, number_of_samples, min_unique_samples, step_sizes,
            get_step_size)


def validate_op2_on_1_permutations(
        op2: Op2 | Callable[[Permutations], Op2],
        search_space: Permutations,
        number_of_samples: int | None = None,
        min_unique_samples: int | Callable[[
            int, Permutations], int] | None = None) -> None:
    """
    Validate the binary operator on one `Permutations` instance.

    :param op2: the operator or operator factory
    :param search_space: the search space
    :param number_of_samples: the optional number of samples
    :param min_unique_samples: the optional unique samples
    """
    args: dict[str, Any] = {
        "op2": op2(search_space) if callable(op2) else op2,
        "search_space": search_space,
        "make_search_space_element_valid":
            make_permutation_valid(search_space),
    }
    if number_of_samples is not None:
        args["number_of_samples"] = number_of_samples
    if min_unique_samples is not None:
        args["min_unique_samples"] = min_unique_samples
    validate_op2(**args)


def validate_op2_on_permutations(
        op2: Op2 | Callable[[Permutations], Op2],
        number_of_samples: int | None = None,
        min_unique_samples: int | Callable[[
            int, Permutations], int] | None = None,
        perm_filter: Callable[[Permutations], bool] | None = None) -> None:
    """
    Validate the binary operator on several `Permutations` instances.

    :param op2: the operator or operator factory
    :param number_of_samples: the optional number of samples
    :param min_unique_samples: the optional unique samples
    :param perm_filter: an optional filter to sort out permutations we cannot
        use for testing
    """
    for pwr in permutations_for_tests(perm_filter):
        validate_op2_on_1_permutations(op2, pwr, number_of_samples,
                                       min_unique_samples)



"""
An objective function counting the number of ones in a bit string.

1. Heinz Mühlenbein. How Genetic Algorithms Really Work: Mutation and
   Hillclimbing. In Reinhard Männer and Bernard Manderick, editors,
   *Proceedings of Parallel Problem Solving from Nature 2 (PPSN-II),*
   September 28-30, 1992, Brussels, Belgium, pages 15-26. Elsevier.
   https://www.researchgate.net/publication/220702092
2. Stefan Droste, Thomas Jansen, and Ingo Wegener. Upper and Lower Bounds for
   Randomized Search Heuristics in Black-Box Optimization. *Theory of
   Computing Systems.* 39(4):525-544. July 2006.
   doi: https://doi.org/10.1007/s00224-004-1177-z
"""

import numba  # type: ignore
import numpy as np

from moptipy.examples.bitstrings.bitstring_problem import BitStringProblem


@numba.njit(nogil=True, cache=True)
def onemax(x: np.ndarray) -> int:
    """
    Get the length of a string minus the number of ones in it.

    :param x: the np array
    :return: the length of the string minus the number of ones, i.e., the
        number of zeros

    >>> print(onemax(np.array([True, True, False, False, False])))
    3
    >>> print(onemax(np.array([True, False, True, False, False])))
    3
    >>> print(onemax(np.array([False, True,  True, False, False])))
    3
    >>> print(onemax(np.array([True, True, True, True, True])))
    0
    >>> print(onemax(np.array([False, True, True, True, True])))
    1
    >>> print(onemax(np.array([False, False, False, False, False])))
    5
    """
    return int(len(x) - x.sum())


class OneMax(BitStringProblem):
    """Maximize the number of ones in a bit string."""

    def __init__(self, n: int) -> None:  # +book
        """
        Initialize the onemax objective function.

        :param n: the dimension of the problem

        >>> print(OneMax(2).n)
        2
        >>> print(OneMax(4).evaluate(np.array([True, True, False, True])))
        1
        """
        super().__init__(n)
        self.evaluate = onemax  # type: ignore

    def __str__(self) -> str:
        """
        Get the name of the onemax objective function.

        :return: `onemax_` + length of string

        >>> print(OneMax(13))
        onemax_13
        """
        return f"onemax_{self.n}"



"""The base class with the information of a build."""

import codecs
import os.path
from io import TextIOBase
from re import MULTILINE
from re import compile as _compile
from typing import Final, Iterable, Pattern, cast

from moptipy.utils.strings import regex_sub
from moptipy.utils.types import type_error


def _canonicalize_path(path: str) -> str:
    """
    Check and canonicalize a path.

    :param path: the path
    :return: the canonicalized path
    """
    if not isinstance(path, str):
        raise type_error(path, "path", str)
    if len(path) <= 0:
        raise ValueError("Path must not be empty.")

    path = os.path.normcase(
        os.path.abspath(
            os.path.realpath(
                os.path.expanduser(
                    os.path.expandvars(path)))))
    if not isinstance(path, str):
        raise type_error(path, "canonicalized path", str)
    if len(path) <= 0:
        raise ValueError("Canonicalization must yield non-empty string, "
                         f"but returned {path!r}.")
    if path in [".", ".."]:
        raise ValueError(f"Canonicalization cannot yield {path!r}.")
    return path


#: the UTF-8 encoding
UTF8: Final[str] = "utf-8-sig"

#: The list of possible text encodings
__ENCODINGS: Final[tuple[tuple[tuple[bytes, ...], str], ...]] = \
    (((codecs.BOM_UTF8,), UTF8),
     ((codecs.BOM_UTF32_LE, codecs.BOM_UTF32_BE), "utf-32"),
     ((codecs.BOM_UTF16_LE, codecs.BOM_UTF16_BE), "utf-16"))


def _get_text_encoding(filename: str) -> str:
    """
    Get the text encoding from a BOM if present.

    Adapted from https://stackoverflow.com/questions/13590749.

    :param filename: the filename
    :return: the encoding
    """
    with open(filename, "rb") as f:
        header = f.read(4)  # Read just the first four bytes.
    for boms, encoding in __ENCODINGS:
        for bom in boms:
            if header.find(bom) == 0:
                return encoding
    return UTF8


#: a pattern used to clean up training white space
_PATTERN_TRAILING_WHITESPACE: Final[Pattern] = \
    _compile(r"[ \t]+\n", MULTILINE)


class Path(str):
    """An immutable representation of a path."""

    def __new__(cls, value):  # noqa
        """
        Construct the object.

        :param value: the string value
        """
        return super().__new__(cls, _canonicalize_path(value))

    def enforce_file(self) -> None:
        """
        Enforce that the path references an existing file.

        :raises ValueError:  if `path` does not reference an existing file
        """
        if not self.is_file():
            raise ValueError(f"Path {self!r} does not identify a file.")

    def is_file(self) -> bool:
        """
        Check if this path identifies an existing file.

        :returns: True if this path identifies an existing file, False
            otherwise.
        """
        return os.path.isfile(self)

    def enforce_dir(self) -> None:
        """
        Enforce that the path references an existing directory.

        :raises ValueError:  if `path` does not reference an existing directory
        """
        if not os.path.isdir(self):
            raise ValueError(f"Path {self!r} does not identify a directory.")

    def contains(self, other: str) -> bool:
        """
        Check whether this path is a directory and contains another path.

        :param other: the other path
        :return: `True` is this path contains the other path, `False` of not
        """
        return os.path.isdir(self) and (
            os.path.commonpath([self])
            == os.path.commonpath([self, Path.path(other)]))

    def enforce_contains(self, other: str) -> None:
        """
        Raise an exception if this is not a directory containing another path.

        :param other: the other path
        :raises ValueError: if `other` is not a sub-path of this path
        """
        self.enforce_dir()
        if not self.contains(other):
            raise ValueError(f"Path {self!r} does not contain {other!r}.")

    def resolve_inside(self, relative_path: str) -> "Path":
        """
        Resolve a relative path to an absolute path inside this path.

        :param relative_path: the path to resolve
        :return: the resolved child path
        :raises ValueError: If the path would resolve to something outside of
            this path and/or if it is empty.
        """
        relative_path = relative_path.strip()
        if not relative_path:
            raise ValueError("Relative path cannot empty.")
        opath: Final[Path] = Path.path(os.path.join(self, relative_path))
        self.enforce_contains(opath)
        return opath

    def ensure_file_exists(self) -> bool:
        """
        Atomically ensure that the file exists and create it otherwise.

        :return: `True` if the file already existed and
            `False` if it was newly and atomically created.
        :raises: ValueError if anything goes wrong during the file creation
        """
        existed: bool = False
        try:
            os.close(os.open(self, os.O_CREAT | os.O_EXCL))
        except FileExistsError:
            existed = True
        except Exception as err:
            if isinstance(err, ValueError):
                raise
            raise ValueError(
                f"Error when trying to create file {self!r}.") from err
        self.enforce_file()
        return existed

    def ensure_dir_exists(self) -> None:
        """Make sure that the directory exists, create it otherwise."""
        try:
            os.makedirs(name=self, exist_ok=True)
        except FileExistsError:
            pass
        except Exception as err:
            if isinstance(err, ValueError):
                raise
            raise ValueError(
                f"Error when trying to create directory {self!r}.") from err
        self.enforce_dir()

    def open_for_read(self) -> TextIOBase:
        """
        Open this file for reading.

        :return: the file open for reading
        """
        return cast(TextIOBase, open(  # noqa
            self, encoding=_get_text_encoding(self),
            errors="strict"))

    def read_all_list(self) -> list[str]:
        """
        Read all the lines in a file.

        :return: the list of strings of text
        """
        self.enforce_file()
        with self.open_for_read() as reader:
            ret = reader.readlines()
        if not isinstance(ret, list):
            raise type_error(ret, f"return value of reading {self!r}", list)
        if len(ret) <= 0:
            raise ValueError(f"File {self!r} contains no text.")
        return [s.rstrip() for s in ret]

    def read_all_str(self) -> str:
        """
        Read a file as a single string.

        :return: the single string of text
        """
        self.enforce_file()
        with self.open_for_read() as reader:
            ret = reader.read()
        if not isinstance(ret, str):
            raise type_error(ret, f"return value of reading {self!r}", str)
        if len(ret) <= 0:
            raise ValueError(f"File {self!r} contains no text.")
        return ret

    def open_for_write(self) -> TextIOBase:
        """
        Open the file for writing.

        :return: the text io wrapper for writing
        """
        return cast(TextIOBase, open(  # noqa
            self, mode="w", encoding="utf-8", errors="strict"))

    def write_all(self, contents: str | Iterable[str]) -> None:
        """
        Write all the lines to this file.

        :param contents: the contents to write
        """
        self.ensure_file_exists()
        if not isinstance(contents, str | Iterable):
            raise type_error(contents, "contents", (str, Iterable))
        all_text = contents if isinstance(contents, str) \
            else "\n".join(contents)
        if len(all_text) <= 0:
            raise ValueError("Writing empty text is not permitted.")
        all_text = regex_sub(_PATTERN_TRAILING_WHITESPACE,
                             "\n", all_text.rstrip())
        if len(all_text) <= 0:
            raise ValueError(
                "Text becomes empty after removing trailing whitespace?")
        with self.open_for_write() as writer:
            writer.write(all_text)
            if all_text[-1] != "\n":
                writer.write("\n")

    def create_file_or_truncate(self) -> None:
        """
        Create the file identified by this path and truncate it it it exists.

        :raises: ValueError if anything goes wrong during the file creation
        """
        try:
            os.close(os.open(self, os.O_CREAT | os.O_TRUNC))
        except FileExistsError as err:
            raise ValueError(
                f"File {self!r} could not be truncated.") from err
        except Exception as err:
            raise ValueError(
                f"Error when trying to create  file {self!r}.") from err
        return self.enforce_file()

    @staticmethod
    def path(path: str) -> "Path":
        """
        Get a canonical path.

        :param path: the path to canonicalize
        :return: the `Path` instance
        """
        if isinstance(path, Path):
            return cast(Path, path)
        return Path(path)

    @staticmethod
    def file(path: str) -> "Path":
        """
        Get a path identifying a file.

        :param path: the path
        :return: the file
        """
        fi: Final[Path] = Path.path(path)
        fi.enforce_file()
        return fi

    @staticmethod
    def directory(path: str) -> "Path":
        """
        Get a path identifying a directory.

        :param path: the path
        :return: the file
        """
        fi: Final[Path] = Path.path(path)
        fi.enforce_dir()
        return fi



"""
The Jump problem.

The jump problem is basically OneMax, but with a deceptive region of k bit
flips before the optimum.

1. Stefan Droste, Thomas Jansen, and Ingo Wegener. On the Analysis of the
   (1+1) Evolutionary Algorithm. *Theoretical Computer Science.*
   276(1-2):51-81. April 2002.
   doi: https://doi.org/10.1016/S0304-3975(01)00182-7
2. Tobias Friedrich, Francesco Quinzan, and Markus Wagner. Escaping Large
   Deceptive Basins of Attraction with Heavy-Tailed Mutation Operators. In
   Hernán E. Aguirre and Keiki Takadama, editors, *Proceedings of the Genetic
   and Evolutionary Computation Conference (GECCO'18),* July 15-19, 2018.
   Kyoto, Japan. ACM. doi: https://doi.org/10.1145/3205455.3205515}
3. Francesco Quinzan, Andreas Göbel, Markus Wagner, and Tobias Friedrich.
   Evolutionary algorithms and submodular functions: benefits of heavy-tailed
   mutations. *Natural Computing.* 20(3):561-575. September 2021.
   doi: https://doi.org/10.1007/s11047-021-09841-7.
   Also: arXiv:1805.10902v2 [cs.DS] 21 Nov 2018.
   https://arxiv.org/abs/1805.10902
"""

from typing import Final

import numba  # type: ignore
import numpy as np

from moptipy.examples.bitstrings.bitstring_problem import BitStringProblem
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.types import check_int_range


@numba.njit(nogil=True, cache=True)
def jump(x: np.ndarray, k: int) -> int:
    """
    Compute the jump value.

    :param x: the np array
    :param k: the k parameter
    :return: jump value

    >>> print(jump(np.array([False, False, False, False, False, False]), 2))
    6
    >>> print(jump(np.array([False, False, False, False, True, False]), 2))
    5
    >>> print(jump(np.array([False, True, True, False, False, False]), 2))
    4
    >>> print(jump(np.array([True, False, True, False, True, False]), 2))
    3
    >>> print(jump(np.array([True, False, True, False, True, True]), 2))
    2
    >>> print(jump(np.array([True, True, True, True, True, False]), 2))
    7
    >>> print(jump(np.array([True, True, True, True, True, True]), 2))
    0
    """
    res: Final[int] = x.sum()
    n: Final[int] = len(x)
    nmk: Final[int] = n - k
    if (res >= n) or (res <= nmk):
        return int(n - res)
    return int(k + res)


class Jump(BitStringProblem):
    """Compute the Jump problem."""

    def __init__(self, n: int, k: int) -> None:  # +book
        """
        Initialize the jump objective function.

        :param n: the dimension of the problem
        :param k: the jump length
        """
        super().__init__(n)
        #: the jump width
        self.k: Final[int] = check_int_range(k, "k", 2, (n >> 1) - 1)

    def __str__(self) -> str:
        """
        Get the name of the jump objective function.

        :return: `jump_` + length of string + `_` + k

        >>> print(Jump(13, 4))
        jump_13_4
        """
        return f"jump_{self.n}_{self.k}"

    def evaluate(self, x: np.ndarray) -> int:
        """
        Evaluate a solution to the jump problem.

        :param x: the bit string to evaluate
        :returns: the value of the jump problem for the string
        """
        return jump(x, self.k)

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of this component as key-value pairs.

        :param logger: the logger for the parameters

        >>> from moptipy.utils.logger import InMemoryLogger
        >>> with InMemoryLogger() as l:
        ...     with l.key_values("C") as kv:
        ...         Jump(23, 7).log_parameters_to(kv)
        ...     text = l.get_log()
        >>> text[1]
        'name: jump_23_7'
        >>> text[3]
        'lowerBound: 0'
        >>> text[4]
        'upperBound: 29'
        >>> text[5]
        'n: 23'
        >>> text[6]
        'k: 7'
        >>> len(text)
        8
        """
        super().log_parameters_to(logger)
        logger.key_value("k", self.k)

    def upper_bound(self) -> int:
        """
        Get the upper bound of the jump problem.

        :return: the length of the bit string

        >>> print(Jump(15, 4).upper_bound())
        18
        """
        return self.n + self.k - 1



"""Strings that carry format information."""

from math import inf, isnan, nan
from typing import Final

from moptipy.utils.types import check_int_range, type_error

#: the formatted string represents normal text
TEXT: Final[int] = 0
#: the formatted string represents a number
NUMBER: Final[int] = 1
#: the formatted string represents NaN, i.e., "not a number"
NAN: Final[int] = 2
#: the formatted string represents positive infinity
POSITIVE_INFINITY: Final[int] = 3
#: the formatted string represents negative infinity
NEGATIVE_INFINITY: Final[int] = 4
#: the formatted string represents a special character
SPECIAL: Final[int] = 5


class FormattedStr(str):
    """
    A subclass of `str` capable of holding formatting information.

    This is a version of string that also stores format information that can,
    for example, be used by a text format driver when typesetting text.
    Instances of this class can also be used as normal strings and be printed
    to the console without any issue. However, they also hold information
    about whether the text should be :attr:`bold`, :attr:`italic`, or rendered
    in a monospace :attr:`code` font.

    Furthermore, if a number or numerical string is represented as a formatted
    string, the field :attr:`mode` will be non-zero. If it is `TEXT=1`, the
    string is a normal number, if it is `NAN=2`, the string is "nan", if it is
    `POSITIVE_INFINITY=3`, then the string is `inf`, and if it is
    `NEGATIVE_INFINITY=4`, the string is `-inf`. These values permit a text
    driver, for example, to replace the special numeric values with unicode
    constants or certain commands. It may also choose to replace floating
    point number of the form `1.5E23` with something like `1.5*10^23`. It can
    do so because a non-zero :attr:`mode` indicates that the string is
    definitely representing a number and that numbers in the string did not
    just occur for whatever other reason.

    If the field :attr:`mode` has value `SPECIAL=5`, then the text contains a
    special sequence, e.g., a unicode character outside of the normal range.
    Then, the text format driver can render this character as a special
    entity.
    """

    #: should this string be formatted in bold face?
    bold: bool
    #: should this string be formatted in italic face?
    italic: bool
    #: should this string be formatted in code face?
    code: bool
    #: the special mode: `TEXT`, `NUMBER`, `NAN`, `POSITIVE_INFINITY`,
    #: `NEGATIVE_INFINITY`, or `SPECIAL`
    mode: int

    def __new__(cls, value, bold: bool = False, italic: bool = False,  # noqa
                code: bool = False, mode: int = TEXT):  # noqa
        """
        Construct the object.

        :param value: the string value
        :param bold: should the format be bold face?
        :param italic: should the format be italic face?
        :param code: should the format be code face?
        :param mode: the mode of the formatted string
        """
        if not isinstance(value, str):
            raise type_error(value, "value", str)
        if not isinstance(bold, bool):
            raise type_error(bold, "bold", bool)
        if not isinstance(italic, bool):
            raise type_error(italic, "italic", bool)
        if not isinstance(code, bool):
            raise type_error(code, "code", bool)
        check_int_range(mode, "mode", TEXT, SPECIAL)
        if bold or italic or code or (mode != TEXT):
            ret = super().__new__(cls, value)
            ret.bold = bold
            ret.italic = italic
            ret.code = code
            ret.mode = mode
            return ret
        return value

    @staticmethod
    def add_format(s: str, bold: bool = False, italic: bool = False,
                   code: bool = False) -> str:
        """
        Add the given format to the specified string.

        :param s: the string
        :param bold: should the format be bold face?
        :param italic: should the format be italic face?
        :param code: should the format be code face?

        >>> from typing import cast
        >>> st = "abc"
        >>> type(st)
        <class 'str'>
        >>> fs = cast(FormattedStr, FormattedStr.add_format(st, bold=True))
        >>> type(fs)
        <class 'moptipy.utils.formatted_string.FormattedStr'>
        >>> fs.bold
        True
        >>> fs.italic
        False
        >>> fs = cast(FormattedStr, FormattedStr.add_format(fs, italic=True))
        >>> fs.bold
        True
        >>> fs.italic
        True
        >>> fs.mode
        0
        """
        if isinstance(s, FormattedStr):
            bold = bold or s.bold
            italic = italic or s.italic
            code = code or s.code
            if (bold != s.bold) or (italic != s.italic) or (code != s.code):
                return FormattedStr(s, bold, italic, code, s.mode)
            return s
        if not isinstance(s, str):
            raise type_error(s, "s", str)
        if bold or italic or code:
            return FormattedStr(s, bold, italic, code, TEXT)
        return s

    @staticmethod
    def number(number: int | float | str | None) -> "FormattedStr":
        """
        Create a formatted string representing a number.

        :param number: the original number or numeric string
        :return: the formatted string representing it

        >>> FormattedStr.number(inf)
        'inf'
        >>> FormattedStr.number(inf) is _PINF
        True
        >>> FormattedStr.number(inf).mode
        3
        >>> FormattedStr.number(-inf)
        '-inf'
        >>> FormattedStr.number(-inf) is _NINF
        True
        >>> FormattedStr.number(-inf).mode
        4
        >>> FormattedStr.number(nan)
        'nan'
        >>> FormattedStr.number(nan) is _NAN
        True
        >>> FormattedStr.number(nan).mode
        2
        >>> FormattedStr.number(123)
        '123'
        >>> FormattedStr.number(123e3)
        '123000.0'
        >>> FormattedStr.number(123e3).mode
        1
        """
        if not isinstance(number, int | str):
            if isinstance(number, float):
                if isnan(number):
                    return _NAN
                if number >= inf:
                    return _PINF
                if number <= -inf:
                    return _NINF
            else:
                raise type_error(number, "number", (float, int, str))
        return FormattedStr(str(number), False, False, False, NUMBER)

    @staticmethod
    def special(text: str) -> "FormattedStr":
        r"""
        Create a special string.

        A special string has mode `SPECIAL` set and notifies the text format
        that special formatting conversion, e.g., a translation from the
        character "alpha" (\u03b1) to `\alpha` is required.

        :param text: the text

        >>> FormattedStr.special("bla")
        'bla'
        >>> FormattedStr.special("bla").mode
        5
        """
        return FormattedStr(text, False, False, False, SPECIAL)


#: the constant for not-a-number
_NAN: Final[FormattedStr] = FormattedStr(str(nan), False, False, False, NAN)
#: the constant for positive infinity
_PINF: Final[FormattedStr] = FormattedStr(str(inf), False, False, False,
                                          POSITIVE_INFINITY)
#: the constant for negative infinity
_NINF: Final[FormattedStr] = FormattedStr(str(-inf), False, False, False,
                                          NEGATIVE_INFINITY)



"""A unary operator flipping exactly one bit."""
from typing import Final

import numpy as np
from numpy.random import Generator

from moptipy.api.operators import Op1


class Op1Flip1(Op1):
    """A unary search operation that flips exactly one bit."""

    def op1(self, random: Generator, dest: np.ndarray, x: np.ndarray) -> None:
        """
        Copy `x` into `dest` and flip exactly one bit.

        :param self: the self pointer
        :param random: the random number generator
        :param dest: the destination array to receive the new point
        :param x: the existing point in the search space
        """
        np.copyto(dest, x)  # copy source to destination
        idx: Final[int] = random.integers(len(x))  # get bit index
        dest[idx] = not dest[idx]  # flip bit

    def __str__(self) -> str:
        """
        Get the name of this unary operator.

        :return: "flip1"
        """
        return "flip1"



"""Some internal helper functions."""

from typing import Final

import numba  # type: ignore
import numpy as np

import moptipy.api.logging as lg

#: the maximum FEs of a black-box process
_FULL_KEY_MAX_FES: Final[str] = f"{lg.SCOPE_PROCESS}.{lg.KEY_MAX_FES}"
#: the maximum runtime in milliseconds of a black-box process
_FULL_KEY_MAX_TIME_MILLIS: Final[str] = \
    f"{lg.SCOPE_PROCESS}.{lg.KEY_MAX_TIME_MILLIS}"
#: the goal objective value of a black-box process
_FULL_KEY_GOAL_F: Final[str] = f"{lg.SCOPE_PROCESS}.{lg.KEY_GOAL_F}"
#: the random seed
_FULL_KEY_RAND_SEED: Final[str] = f"{lg.SCOPE_PROCESS}.{lg.KEY_RAND_SEED}"
#: the FE when the best objective value was reached


def _check_max_time_millis(max_time_millis: int | float,
                           total_fes: int | float,
                           total_time_millis: int | float) -> None:
    """
    Check whether a max-time-millis value is permissible.

    If we set a time limit for a run, then the
    :meth:`~moptipy.api.process.Process.should_terminate` will become `True`
    approximately after the time limit has expired. However, this also
    could be later (or maybe even earlier) due to the workings of the
    underlying operating system. And even if
    :meth:`~moptipy.api.process.Process.should_terminate` if `True`, it is not
    clear whether the optimization algorithm can query it right away.
    Instead, it may be blocked in a long-running objective function
    evaluation or some other computation. Hence, it may actually stop
    later. So we cannot simply require that
    `total_time_millis <= max_time_millis`, as this is not practically
    enforceable. Instead, we will heuristically determine a feasible
    maximum limit for how much longer an algorithm might run.

    :param max_time_millis: the max time millis threshold
    :param total_fes: the total FEs performed
    :param total_time_millis: the measured total time millis
    """
    if total_fes == 1:
        return
    div: Final[float] = (total_fes - 2) if total_fes > 1 else 1
    permitted_limit: Final[float] = \
        60_000 + ((1 + (1.1 * (max_time_millis / div))) * total_fes)
    if total_time_millis > permitted_limit:
        raise ValueError(
            f"If max_time_millis is {max_time_millis} and "
            f"total_fes is {total_fes}, then total_time_millis must "
            f"not be more than {permitted_limit}, but is"
            f"{total_time_millis}.")


@numba.njit(nogil=True)
def _get_reach_index(f: np.ndarray, goal_f: int | float):  # noqa
    """
    Compute the offset from the end of `f` when `goal_f` was reached.

    :param f: the raw data array, which must be sorted in
        descending order
    :param goal_f: the goal f value
    :return: the index, or `0` if `goal_f` was not reached

    >>> ft = np.array([10, 9, 8, 5, 3, 2, 1])
    >>> _get_reach_index(ft, 11)
    7
    >>> ft[ft.size - _get_reach_index(ft, 11)]
    10
    >>> _get_reach_index(ft, 10)
    7
    >>> _get_reach_index(ft, 9)
    6
    >>> ft[ft.size - _get_reach_index(ft, 6)]
    5
    >>> _get_reach_index(ft, 1)
    1
    >>> _get_reach_index(ft, 0.9)
    0
    """
    return np.searchsorted(f[::-1], goal_f, side="right")
